Package: blockr.dag
Title: A Directed Acyclic Graph Extension for 'blockr'
Version: 0.1.2
Authors@R: 
    c(person(given = "David",
             family = "Granjon",
             role = c("aut", "cre"),
             email = "david@cynkra.com"),
      person(given = "Nicolas",
             family = "Bennett",
             role = c("aut"),
             email = "nicolas@cynkra.com"),
      person(given = "Christoph",
             family = "Sax",
             role = c("aut"),
             email = "christoph@cynkra.com"),
      person(family = "Bristol Myers Squibb",
             role = "fnd")
    )
Description: Building on the docking layout manager provided by 'blockr.dock', this
    provides an extension that allows for visualizing and manipulating a 'blockr'
    board using a DAG-based user interface powered by the 'g6R' 
    graph visualisation HTML widget.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: blockr.core (>= 0.1.2), blockr.dock (>= 0.1.1), shiny, g6R (>=
        0.6.0), jsonlite, htmltools
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), roxy.shinylive,
        webshot2, quarto, cyclocomp, shinytest2, colorspace
Config/testthat/edition: 3
URL: https://bristolmyerssquibb.github.io/blockr.dag/
VignetteBuilder: quarto
NeedsCompilation: no
Packaged: 2026-04-29 13:53:15 UTC; davidgranjon
Author: David Granjon [aut, cre],
  Nicolas Bennett [aut],
  Christoph Sax [aut],
  Bristol Myers Squibb [fnd]
Maintainer: David Granjon <david@cynkra.com>
Repository: CRAN
Date/Publication: 2026-04-29 15:20:09 UTC
Built: R 4.5.2; ; 2026-05-01 19:32:13 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as_graph,graph)
S3method(as_graph,list)
S3method(blockr.core::blockr_deser,subboard)
S3method(blockr.core::blockr_ser,subboard)
S3method(board_actions,dag_extension)
S3method(context_menu_items,dag_extension)
S3method(context_menu_items,dock_board)
S3method(context_menu_items,dock_extension)
S3method(context_menu_items,dock_extensions)
S3method(context_menu_items,list)
S3method(extension_block_callback,dag_extension)
S3method(toolbar_items,dag_extension)
S3method(toolbar_items,dock_board)
S3method(toolbar_items,dock_extension)
S3method(toolbar_items,dock_extensions)
S3method(toolbar_items,list)
export(as_graph)
export(context_menu_items)
export(graph_combos)
export(graph_edges)
export(graph_nodes)
export(is_context_menu_entry)
export(is_graph)
export(is_toolbar_item)
export(new_context_menu_entry)
export(new_dag_extension)
export(new_graph)
export(new_toolbar_item)
export(toolbar_items)
import(blockr.core)
import(g6R)
import(shiny)
importFrom(blockr.dock,board_actions)
importFrom(blockr.dock,extension_block_callback)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# blockr.dag

<!-- badges: start -->

[![lifecycle](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![status](https://github.com/BristolMyersSquibb/blockr.dag/actions/workflows/ci.yaml/badge.svg)](https://github.com/BristolMyersSquibb/blockr.dag/actions/workflows/ci.yaml)
[![coverage](https://codecov.io/gh/BristolMyersSquibb/blockr.dag/graph/badge.svg?token=s0zz3En4x1)](https://app.codecov.io/gh/BristolMyersSquibb/blockr.dag)
[![CRAN
status](https://www.r-pkg.org/badges/version/blockr.dag)](https://CRAN.R-project.org/package=blockr.dag)
<!-- badges: end -->

An interative network library provided by g6R can be used as front-end
to a blockr board using this package.

## Installation

You can install the development version of blockr.dag from
[GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("BristolMyersSquibb/blockr.dag")
```

## Example

To start up a board with the `dag` extension, run the following code:

``` r
library(blockr.dag)
library(blockr.core)
library(blockr.dock)

options(
  "g6R.mode" = "dev",
  #"g6R.layout_on_data_change" = TRUE,
  "g6R.preserve_elements_position" = TRUE
)

serve(
  new_dock_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_scatter_block(x = "Sepal.Length", y = "Sepal.Width")
    ),
    links = list(from = "a", to = "b", input = "data"),
    stacks = c(
      stack_1 = new_dock_stack(c("a", "b"), color = "#0000FF"),
      stack_2 = new_dock_stack()
    ),
    extensions = new_dag_extension()
  )
)
```

<figure>
<img src="man/figures/demo-app.png"
alt="blockr.dag demo application with stacks" />
<figcaption aria-hidden="true">blockr.dag demo application with
stacks</figcaption>
</figure>

To start up the dag extension with dummy nodes and edges:

``` r
library(blockr.dag)
library(blockr.dock)
library(blockr.core)

graph <- new_graph(
  nodes = list(
    list(id = 1, style = list(labelText = "Node 1")),
    list(id = 2, style = list(labelText = "Node 2"))
  ),
  edges = list(
    list(
      source = 1,
      target = 2,
      style = list(
        labelText = "Edge from 1 to 2"
      )
    )
  )
)

serve(
  new_dock_board(
    extensions = new_dag_extension(graph)
  )
)
```

library(blockr.dag)
library(blockr.core)
library(blockr.dock)

options(
  "g6R.mode" = "dev" #,
  #"g6R.layout_on_data_change" = TRUE
)

serve(
  new_dock_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_head_block(n = 10),
      c = new_subset_block(),
      d = new_head_block(n = 5),
      e = new_rbind_block(),
      f = new_subset_block(),
      g = new_rbind_block(),
      h = new_head_block(n = 3),
      i = new_scatter_block(x = "Sepal.Length", y = "Sepal.Width")
    ),
    links = c(
      new_link(from = "a", to = "b", input = "data"),
      new_link(from = "b", to = "c", input = "data"),
      new_link(from = "c", to = "d", input = "data"),
      new_link(from = "b", to = "f", input = "data"),
      new_link(from = "d", to = "e", input = "1"),
      new_link(from = "f", to = "e", input = "2"),
      new_link(from = "e", to = "g", input = "1"),
      new_link(from = "f", to = "g", input = "2"),
      new_link(from = "g", to = "h", input = "data"),
      new_link(from = "h", to = "i", input = "data")
    ),
    stacks = c(
      new_stack(id = "stack1", blocks = c("b", "c", "d")),
      new_stack(id = "stack2", blocks = c("f", "g"))
    ),
    extensions = new_dag_extension()
  )
)

library(blockr.dag)
library(blockr.core)
library(blockr.dock)

options(
  "g6R.mode" = "dev",
  "g6R.directed_graph" = TRUE,
  "g6R.layout_on_data_change" = TRUE
)

serve(
  id = "board",
  new_dock_board(
    extensions = new_dag_extension()
  )
)

library(blockr.dag)
library(blockr.core)
library(blockr.dock)

options(
  "g6R.mode" = "dev",
  #"g6R.layout_on_data_change" = TRUE,
  "g6R.preserve_elements_position" = TRUE
)

serve(
  new_dock_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_scatter_block(x = "Sepal.Length", y = "Sepal.Width")
    ),
    links = list(from = "a", to = "b", input = "data"),
    stacks = c(
      stack_1 = new_dock_stack(c("a", "b"), color = "#0000FF"),
      stack_2 = new_dock_stack()
    ),
    extensions = new_dag_extension()
  )
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)


## ----setup--------------------------------------------------------------------
library(blockr.dag)
library(blockr.core)
library(blockr.dock)


## -----------------------------------------------------------------------------
# Create an empty board.
serve(new_dock_board(extensions = new_dag_extension()))


## -----------------------------------------------------------------------------
graph <- new_graph(
  nodes = list(
    list(id = 1),
    list(id = 2)
  ),
  edges = list(
    list(
      source = 1,
      target = 2,
      style = list(labelText = "1-2")
    )
  )
)

serve(
  new_dock_board(
    extensions = new_dag_extension(graph)
  )
)


## graph TD
##     subgraph Core Layer
##         blockr.core[blockr.core]
##     end
## 
##     subgraph Extension Blocks
##         blockr.dplyr[blockr.dplyr]
##         blockr.ggplot[blockr.ggplot]
##     end
## 
##     subgraph UI Layer
##         blockr.dock[blockr.dock]
##         blockr.dag[blockr.dag]
##     end
## 
##     subgraph External Dependencies
##         dockViewR[dockViewR]
##         g6R[g6R]
##     end
## 
##     blockr.core --> blockr.dock
##     blockr.dock --> blockr.dag
##     dockViewR --> blockr.dock
##     g6R --> blockr.dag
## 
##     blockr.core --> blockr.dplyr
##     blockr.core --> blockr.ggplot
## 
##     blockr.dplyr -.-> blockr.dock
##     blockr.dplyr -.-> blockr.dag
##     blockr.ggplot -.-> blockr.dock
##     blockr.ggplot -.-> blockr.dag

## -----------------------------------------------------------------------------
# Update observer handles board changes
update_observer <- function(update, board, proxy) {
  observeEvent(update(), {
    upd <- update()

    if (length(upd$blocks$add)) {
      add_nodes(upd$blocks$add, board$board, proxy)
    }

    if (length(upd$stacks$mod)) {
      update_combos(upd$stacks$mod, board$board, proxy)
    }

    # Handle other update types...
  })
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)


## ----setup--------------------------------------------------------------------
library(blockr.dag)


## -----------------------------------------------------------------------------
# Context menu entry
entry <- new_context_menu_entry(
  name = "Custom Action",
  js = function(ns) "alert('Custom action')",
  action = function(board, update, ...) {
    # Server-side logic
  },
  condition = function(board, target) {
    target$type == "node"
  }
)


## -----------------------------------------------------------------------------
# Add custom context menu items to an extension
context_menu_items.dag_extension <- function(x) {
  list(
    new_context_menu_entry(
      name = "My Action",
      js = function(ns) {
        sprintf("Shiny.setInputValue('%s', true)", ns("my_action"))
      },
      action = my_custom_action("my_action"),
      condition = function(board, target) TRUE
    )
  )
}

library(blockr.dag)
library(blockr.dock)
library(blockr.core)

graph <- new_graph(
  nodes = list(
    list(id = 1, style = list(labelText = "Node 1")),
    list(id = 2, style = list(labelText = "Node 2"))
  ),
  edges = list(
    list(
      source = 1,
      target = 2,
      style = list(
        labelText = "Edge from 1 to 2"
      )
    )
  )
)

serve(
  new_dock_board(
    extensions = new_dag_extension(graph)
  )
)

#' Graph object
#'
#' R list-based representation of a g6 graph object.
#'
#' @param nodes Graph nodes (i.e. board blocks).
#' @param edges Graph edges (i.e. board links).
#' @param combos Node groups (i.e. board stacks).
#' @param x Object to test or convert.
#' @param ... Generic consistency.
#'
#' @details
#' \describe{
#'   \item{`new_graph()`}{Creates a new graph
#' object with the specified nodes, edges, and combos.}
#'   \item{`is_graph()`}{Tests whether an object
#' is a valid graph object.}
#'   \item{`graph_nodes()`}{Extracts the nodes
#' component from a graph object.}
#'   \item{`graph_edges()`}{Extracts the edges
#' component from a graph object.}
#'   \item{`graph_combos()`}{Extracts the combos
#' component from a graph object.}
#'   \item{`as_graph()`}{Generic function to
#' convert objects to graph format.}
#' }
#'
#' @rdname graph
#' @export
#' @return
#' \describe{
#'   \item{`new_graph()`}{A graph object
#' of class "graph" containing nodes, edges, and combos lists.}
#'   \item{`is_graph()`}{`TRUE` if `x` is a
#' graph object, `FALSE` otherwise.}
#'   \item{`graph_nodes()`}{A list of graph nodes.}
#'   \item{`graph_edges()`}{A list of graph edges.}
#'   \item{`graph_combos()`}{A list of graph
#' combos (node groups).}
#'   \item{`as_graph()`}{A graph object converted
#' from the input.}
#' }
new_graph <- function(nodes = list(), edges = list(), combos = list()) {
  structure(
    list(nodes = nodes, edges = edges, combos = combos),
    class = "graph"
  )
}

#' @rdname graph
#' @export
is_graph <- function(x) {
  inherits(x, "graph")
}

#' @rdname graph
#' @export
graph_nodes <- function(x) {
  x[["nodes"]]
}

#' @rdname graph
#' @export
graph_edges <- function(x) {
  x[["edges"]]
}

#' @rdname graph
#' @export
graph_combos <- function(x) {
  x[["combos"]]
}

#' @rdname graph
#' @export
as_graph <- function(x, ...) {
  UseMethod("as_graph")
}

#' @rdname graph
#' @export
as_graph.graph <- function(x, ...) {
  x
}

#' @rdname graph
#' @export
as_graph.list <- function(x, ...) {
  do.call(new_graph, x)
}

library(testthat)

test_that("new_graph creates valid graph objects", {
  # Test with default empty lists
  g <- new_graph()

  expect_s3_class(g, "graph")
  expect_type(g, "list")
  expect_named(g, c("nodes", "edges", "combos"))
  expect_equal(length(g$nodes), 0)
  expect_equal(length(g$edges), 0)
  expect_equal(length(g$combos), 0)

  # Test with provided data
  nodes <- list(list(id = "1"), list(id = "2"))
  edges <- list(list(source = "1", target = "2"))
  combos <- list(list(id = "combo1"))

  g <- new_graph(nodes, edges, combos)

  expect_s3_class(g, "graph")
  expect_identical(g$nodes, nodes)
  expect_identical(g$edges, edges)
  expect_identical(g$combos, combos)
})

test_that("is_graph correctly identifies graph objects", {
  g <- new_graph()

  expect_true(is_graph(g))
  expect_false(is_graph(list()))
  expect_false(is_graph("not a graph"))
  expect_false(is_graph(NULL))
  expect_false(is_graph(123))
  expect_false(is_graph(data.frame()))
})

test_that("graph_nodes extracts nodes correctly", {
  nodes <- list(list(id = "1"), list(id = "2"))
  g <- new_graph(nodes = nodes)

  expect_identical(graph_nodes(g), nodes)

  # Test with empty graph
  empty_g <- new_graph()
  expect_equal(length(graph_nodes(empty_g)), 0)
  expect_type(graph_nodes(empty_g), "list")
})

test_that("graph_edges extracts edges correctly", {
  edges <- list(
    list(source = "1", target = "2"),
    list(source = "2", target = "3")
  )
  g <- new_graph(edges = edges)

  expect_identical(graph_edges(g), edges)

  # Test with empty graph
  empty_g <- new_graph()
  expect_equal(length(graph_edges(empty_g)), 0)
  expect_type(graph_edges(empty_g), "list")
})

test_that("graph_combos extracts combos correctly", {
  combos <- list(list(id = "combo1"), list(id = "combo2"))
  g <- new_graph(combos = combos)

  expect_identical(graph_combos(g), combos)

  # Test with empty graph
  empty_g <- new_graph()
  expect_equal(length(graph_combos(empty_g)), 0)
  expect_type(graph_combos(empty_g), "list")
})

test_that("as_graph.graph returns identical object", {
  g1 <- new_graph(
    nodes = list(list(id = "1")),
    edges = list(list(source = "1", target = "2"))
  )

  g2 <- as_graph(g1)

  expect_identical(g1, g2)
  expect_s3_class(g2, "graph")
})

test_that("as_graph.list converts list to graph", {
  g_data <- list(
    nodes = list(list(id = "1"), list(id = "2")),
    edges = list(list(source = "1", target = "2")),
    combos = list(list(id = "combo1"))
  )

  g <- as_graph(g_data)

  expect_s3_class(g, "graph")
  expect_identical(graph_nodes(g), g_data$nodes)
  expect_identical(graph_edges(g), g_data$edges)
  expect_identical(graph_combos(g), g_data$combos)
})

test_that("as_graph.list works with partial lists", {
  # Only nodes
  g_data <- list(nodes = list(list(id = "test")))
  g <- as_graph(g_data)

  expect_s3_class(g, "graph")
  expect_equal(length(graph_nodes(g)), 1)
  expect_equal(length(graph_edges(g)), 0)
  expect_equal(length(graph_combos(g)), 0)

  # Only edges
  g_data <- list(edges = list(list(source = "1", target = "2")))
  g <- as_graph(g_data)

  expect_s3_class(g, "graph")
  expect_equal(length(graph_nodes(g)), 0)
  expect_equal(length(graph_edges(g)), 1)
  expect_equal(length(graph_combos(g)), 0)

  # Empty list
  g_data <- list()
  g <- as_graph(g_data)

  expect_s3_class(g, "graph")
  expect_equal(length(graph_nodes(g)), 0)
  expect_equal(length(graph_edges(g)), 0)
  expect_equal(length(graph_combos(g)), 0)
})

test_that("as_graph generic dispatches correctly", {
  # Test that UseMethod works correctly
  g <- new_graph()
  expect_identical(as_graph(g), g)

  g_data <- list(nodes = list(), edges = list(), combos = list())
  g <- as_graph(g_data)
  expect_s3_class(g, "graph")
})

test_that("accessor functions handle missing components gracefully", {
  # Create graph-like object missing some components
  incomplete <- structure(list(nodes = list()), class = "graph")

  expect_equal(graph_nodes(incomplete), list())
  expect_null(graph_edges(incomplete))
  expect_null(graph_combos(incomplete))
})

draw_link_action <- function(trigger, board, update, dag_extension, ...) {
  blockr.dock::new_action(
    function(input, output, session) {
      proxy <- blockr_g6_proxy(session)
      observeEvent(
        trigger(),
        {
          new <- trigger()

          blocks <- board_blocks(board$board)

          inps <- blockr.dock::block_input_select(
            blocks[[new$target]],
            new$target,
            board_links(board$board),
            mode = "inputs"
          )

          if (length(inps) == 0L) {
            notify(
              "No inputs are available for block {new$target}.",
              type = "warning"
            )

            return()
          }

          new_lnk <- new_link(
            from = new$source,
            to = new$target,
            input = inps[1L]
          )

          update(list(links = list(add = as_links(new_lnk))))
        },
        label = "draw_link"
      )

      NULL
    },
    id = "draw_link_action"
  )
}

remove_selected_action <- function(trigger, board, update, dag_extension, ...) {
  blockr.dock::new_action(
    function(input, output, session) {
      input <- dag_extension[["proxy"]]$session$input
      observeEvent(
        trigger(),
        {
          update(
            list(
              blocks = list(
                rm = from_g6_node_id(
                  input[[paste0(graph_id(), "-selected_node")]]
                )
              ),
              links = list(
                rm = from_g6_edge_id(
                  input[[paste0(graph_id(), "-selected_edge")]]
                )
              ),
              stacks = list(
                rm = from_g6_combo_id(
                  input[[paste0(graph_id(), "-selected_combo")]]
                )
              )
            )
          )
        },
        label = "remove_selected"
      )

      NULL
    },
    id = "remove_selected_action"
  )
}

copy_selection_to_clipboard <- function(board, dag_extension) {
  input <- dag_extension[["proxy"]]$session$input
  selected_nodes <- input[[paste0(graph_id(), "-selected_node")]]
  selected_combos <- input[[paste0(graph_id(), "-selected_combo")]]

  subboard <- extract_subboard(
    board$board,
    block_ids = from_g6_node_id(coal(selected_nodes, character())),
    stack_ids = from_g6_combo_id(coal(selected_combos, character()))
  )

  if (is.null(subboard)) {
    return(NULL)
  }

  states <- live_block_states(board, names(subboard$blocks))

  session <- dag_extension[["proxy"]]$session

  json <- tryCatch(
    as.character(jsonlite::toJSON(
      blockr_ser(subboard, blocks = states),
      auto_unbox = TRUE
    )),
    error = function(e) {
      showNotification(
        paste("Cannot copy selection:", conditionMessage(e)),
        type = "error"
      )
      NULL
    }
  )

  if (is.null(json)) {
    return(NULL)
  }

  session$sendCustomMessage("write-clipboard", list(json = json))

  subboard
}

remove_subboard <- function(subboard, update) {
  update(list(
    blocks = list(rm = names(subboard$blocks)),
    links = list(rm = names(subboard$links)),
    stacks = list(rm = names(subboard$stacks))
  ))
}

copy_selected_action <- function(trigger, board, update, dag_extension, ...) {
  blockr.dock::new_action(
    function(input, output, session) {
      observeEvent(trigger(), {
        copy_selection_to_clipboard(board, dag_extension)
      }, label = "copy_selected")
      NULL
    },
    id = "copy_selected_action"
  )
}

cut_selected_action <- function(trigger, board, update, dag_extension, ...) {
  blockr.dock::new_action(
    function(input, output, session) {
      observeEvent(trigger(), {
        subboard <- copy_selection_to_clipboard(board, dag_extension)
        if (!is.null(subboard)) remove_subboard(subboard, update)
      }, label = "cut_selected")
      NULL
    },
    id = "cut_selected_action"
  )
}

paste_action <- function(trigger, board, update, dag_extension, ...) {
  blockr.dock::new_action(
    function(input, output, session) {
      observeEvent(trigger(), {
        subboard <- NULL
        tryCatch(
          subboard <- blockr_deser(jsonlite::fromJSON(
            trigger(),
            simplifyDataFrame = FALSE
          )),
          error = function(e) {
            showNotification(
              paste("Cannot paste clipboard content:", conditionMessage(e)),
              type = "error"
            )
          }
        )
        if (is.null(subboard)) {
          return()
        }

        remapped <- remap_subboard_ids(subboard, board$board)

        update(list(
          blocks = list(add = remapped$blocks),
          links = list(add = remapped$links),
          stacks = list(add = remapped$stacks)
        ))
      }, label = "paste")
      NULL
    },
    id = "paste_action"
  )
}

#' @importFrom blockr.dock board_actions
#' @export
board_actions.dag_extension <- function(x, ...) {
  list(
    draw_link_action,
    remove_selected_action,
    copy_selected_action,
    cut_selected_action,
    paste_action
  )
}

my_board <- new_board(
  blocks = c(
    a = new_dataset_block(),
    b = new_head_block(),
    c = new_dataset_block(),
    d = new_scatter_block()
  ),
  links = list(
    from = "c",
    to = "d",
    input = "data"
  )
)

test_draw_link_action <- function(edge, expected_updates) {
  r_board <- reactiveValues(board = my_board)
  r_update <- reactiveVal(list())

  testServer(
    function(id, ...) {
      moduleServer(
        id,
        module = draw_link_action(
          trigger = reactive(edge),
          board = r_board,
          update = r_update,
          dag_extension = list(
            proxy = g6_proxy(
              "graph",
              session = MockShinySession$new()
            )
          )
        )
      )
    },
    {
      expect_length(r_update(), 0L)
      session$flushReact()

      upd <- r_update()
      expect_length(upd, expected_updates)

      if (expected_updates > 0) {
        expect_named(upd, "links")
        expect_length(upd$links, 1L)
        expect_named(upd$links, "add")
        expect_length(upd$links$add, 1L)

        new_link <- as.list(upd$links$add[[1]])
        expect_identical(new_link$from, "a")
        expect_identical(new_link$to, "b")
        expect_identical(new_link$input, "data")
      }
    }
  )
}

test_that("draw link action with valid target", {
  test_draw_link_action(
    list(id = "a-b", source = "a", target = "b", targetPort = "node-b-data"),
    1L
  )
})

test_that("remove_selected_action works", {
  r_board <- reactiveValues(board = my_board)
  r_update <- reactiveVal(list())

  testServer(
    function(id, ...) {
      moduleServer(
        id,
        module = remove_selected_action(
          trigger = reactive(TRUE),
          board = r_board,
          update = r_update,
          dag_extension = list(
            proxy = g6_proxy(
              "graph",
              session = MockShinySession$new()
            )
          )
        )
      )
    },
    {
      expect_length(r_update(), 0L)
      dag_extension$proxy$session$setInputs(
        "graph-selected_node" = c("a", "b"),
        "graph-selected_edge" = "c-d"
      )
      upd <- r_update()
      expect_length(upd, 3L)
      expect_named(upd, c("blocks", "links", "stacks"))
      expect_named(upd$blocks, "rm")
      expect_identical(upd$blocks$rm, input[["graph-selected_node"]])
      expect_named(upd$links, "rm")
      expect_identical(upd$links$rm, input[["graph-selected_edge"]])
      expect_named(upd$stacks, "rm")
      expect_length(upd$stacks$rm, 0L)
    }
  )
})

# --- Copy / Cut / Paste fixtures and helpers ---

copy_board <- new_board(
  blocks = c(
    a = new_dataset_block(),
    b = new_head_block(),
    c = new_dataset_block(),
    d = new_scatter_block()
  ),
  links = list(from = "a", to = "b", input = "data"),
  stacks = list(s1 = new_stack(c("a", "b")))
)

make_clipboard_spy <- function() {
  spy <- new.env(parent = emptyenv())
  spy$msgs <- list()
  mock_session <- MockShinySession$new()
  mock_session$sendCustomMessage <- function(type, message) {
    spy$msgs[[length(spy$msgs) + 1L]] <- list(type = type, message = message)
  }
  list(session = mock_session, spy = spy)
}

# --- copy_selected_action ---

test_that("copy_selected_action sends clipboard message for selected nodes", {
  cs <- make_clipboard_spy()
  r_board <- reactiveValues(board = copy_board, blocks = list())
  r_update <- reactiveVal(list())

  testServer(
    function(id, ...) {
      moduleServer(
        id,
        module = copy_selected_action(
          trigger = reactive(TRUE),
          board = r_board,
          update = r_update,
          dag_extension = list(
            proxy = g6_proxy("graph", session = cs$session)
          )
        )
      )
    },
    {
      expect_length(r_update(), 0L)
      dag_extension$proxy$session$setInputs(
        "graph-selected_node" = c("node-a", "node-b")
      )
      session$flushReact()

      expect_length(cs$spy$msgs, 1L)
      expect_identical(cs$spy$msgs[[1]]$type, "write-clipboard")

      json_data <- jsonlite::fromJSON(
        cs$spy$msgs[[1]]$message$json,
        simplifyDataFrame = FALSE
      )
      expect_identical(json_data$object, "subboard")

      upd <- r_update()
      expect_length(upd, 0L)
    }
  )
})

test_that("copy_selected_action does nothing when nothing is selected", {
  cs <- make_clipboard_spy()
  r_board <- reactiveValues(board = copy_board, blocks = list())
  r_update <- reactiveVal(list())

  testServer(
    function(id, ...) {
      moduleServer(
        id,
        module = copy_selected_action(
          trigger = reactive(TRUE),
          board = r_board,
          update = r_update,
          dag_extension = list(
            proxy = g6_proxy("graph", session = cs$session)
          )
        )
      )
    },
    {
      expect_length(r_update(), 0L)
      session$flushReact()

      expect_length(cs$spy$msgs, 0L)
      expect_length(r_update(), 0L)
    }
  )
})

# --- cut_selected_action ---

test_that("cut_selected_action copies to clipboard and removes", {
  cs <- make_clipboard_spy()
  r_board <- reactiveValues(board = copy_board, blocks = list())
  r_update <- reactiveVal(list())

  testServer(
    function(id, ...) {
      moduleServer(
        id,
        module = cut_selected_action(
          trigger = reactive(TRUE),
          board = r_board,
          update = r_update,
          dag_extension = list(
            proxy = g6_proxy("graph", session = cs$session)
          )
        )
      )
    },
    {
      expect_length(r_update(), 0L)
      dag_extension$proxy$session$setInputs(
        "graph-selected_node" = c("node-a", "node-b")
      )
      session$flushReact()

      expect_length(cs$spy$msgs, 1L)
      expect_identical(cs$spy$msgs[[1]]$type, "write-clipboard")

      upd <- r_update()
      expect_length(upd, 3L)
      expect_named(upd, c("blocks", "links", "stacks"))
      expect_named(upd$blocks, "rm")
      expect_true(all(c("a", "b") %in% upd$blocks$rm))
      expect_named(upd$links, "rm")
      expect_named(upd$stacks, "rm")
    }
  )
})

test_that("cut_selected_action does nothing when nothing is selected", {
  cs <- make_clipboard_spy()
  r_board <- reactiveValues(board = copy_board, blocks = list())
  r_update <- reactiveVal(list())

  testServer(
    function(id, ...) {
      moduleServer(
        id,
        module = cut_selected_action(
          trigger = reactive(TRUE),
          board = r_board,
          update = r_update,
          dag_extension = list(
            proxy = g6_proxy("graph", session = cs$session)
          )
        )
      )
    },
    {
      expect_length(r_update(), 0L)
      session$flushReact()

      expect_length(cs$spy$msgs, 0L)
      expect_length(r_update(), 0L)
    }
  )
})

# --- paste_action ---

test_that("paste_action adds remapped blocks/links/stacks", {
  sub <- extract_subboard(copy_board, block_ids = c("a", "b"))
  json_str <- as.character(jsonlite::toJSON(
    blockr_ser(sub),
    auto_unbox = TRUE
  ))
  r_board <- reactiveValues(board = copy_board)
  r_update <- reactiveVal(list())

  testServer(
    function(id, ...) {
      moduleServer(
        id,
        module = paste_action(
          trigger = reactive(json_str),
          board = r_board,
          update = r_update,
          dag_extension = list(
            proxy = g6_proxy("graph", session = MockShinySession$new())
          )
        )
      )
    },
    {
      expect_length(r_update(), 0L)
      session$flushReact()

      upd <- r_update()
      expect_named(upd, c("blocks", "links", "stacks"))
      expect_named(upd$blocks, "add")
      expect_named(upd$links, "add")
      expect_named(upd$stacks, "add")

      expect_length(upd$blocks$add, 2L)
      new_block_ids <- names(upd$blocks$add)
      expect_true(!any(new_block_ids %in% board_block_ids(copy_board)))
    }
  )
})

test_that("paste_action ignores invalid JSON", {
  r_board <- reactiveValues(board = copy_board)
  r_update <- reactiveVal(list())

  testServer(
    function(id, ...) {
      moduleServer(
        id,
        module = paste_action(
          trigger = reactive("not valid json {{{"),
          board = r_board,
          update = r_update,
          dag_extension = list(
            proxy = g6_proxy("graph", session = MockShinySession$new())
          )
        )
      )
    },
    {
      expect_length(r_update(), 0L)
      session$flushReact()
      expect_length(r_update(), 0L)
    }
  )
})

test_that("paste_action with empty subboard produces empty add lists", {
  empty_sub <- new_subboard(blocks(), links(), stacks())
  json_str <- as.character(jsonlite::toJSON(
    blockr_ser(empty_sub),
    auto_unbox = TRUE
  ))
  r_board <- reactiveValues(board = copy_board)
  r_update <- reactiveVal(list())

  testServer(
    function(id, ...) {
      moduleServer(
        id,
        module = paste_action(
          trigger = reactive(json_str),
          board = r_board,
          update = r_update,
          dag_extension = list(
            proxy = g6_proxy("graph", session = MockShinySession$new())
          )
        )
      )
    },
    {
      expect_length(r_update(), 0L)
      session$flushReact()

      upd <- r_update()
      expect_named(upd, c("blocks", "links", "stacks"))
      expect_length(upd$blocks$add, 0L)
      expect_length(upd$links$add, 0L)
      expect_length(upd$stacks$add, 0L)
    }
  )
})

# --- remove_subboard() helper ---

test_that("remove_subboard() produces correct rm lists", {
  sb <- extract_subboard(copy_board, block_ids = c("a", "b"))
  captured <- NULL
  mock_update <- function(val) captured <<- val

  remove_subboard(sb, mock_update)

  expect_named(captured, c("blocks", "links", "stacks"))
  expect_identical(captured$blocks$rm, names(sb$blocks))
  expect_identical(captured$links$rm, names(sb$links))
  expect_identical(captured$stacks$rm, names(sb$stacks))
})

graph_id <- function(ns = NULL) {
  res <- "graph"

  if (is.null(ns)) {
    return(res)
  }

  ns(res)
}

last <- function(x) x[[length(x)]]

filter_null <- function(x) Filter(Negate(is.null), x)

has_length <- function(x) length(x) > 0L

# nocov start
preprocess_graph_state <- function() {
  ns <- get_session()$ns
  shiny::snapshotPreprocessInput(
    ns(sprintf("%s-state", graph_id())),
    function(value) {
      strip_positions <- function(elements) {
        lapply(elements, function(el) {
          if (!is.null(el$style)) {
            el$style$x <- NULL
            el$style$y <- NULL
            el$style$z <- NULL
            el$style$src <- "<shinytest2:preprocessed>"
          }
          el
        })
      }

      if (!is.null(value$nodes)) {
        value$nodes <- strip_positions(value$nodes)
      }
      if (!is.null(value$edges)) {
        value$edges <- strip_positions(value$edges)
      }
      if (!is.null(value$combos)) {
        value$combos <- strip_positions(value$combos)
      }
      value
    }
  )
}

preprocess_mouse_position <- function() {
  ns <- get_session()$ns
  shiny::snapshotPreprocessInput(
    ns(sprintf("%s-mouse_position", graph_id())),
    function(value) {
      "<shinytest2:preprocessed>"
    }
  )
} # nocov end

update_action_trigger <- function(action_name, input_name) {
  function(actions, session = get_session()) {
    observeEvent(
      session$input[[input_name]],
      actions[[action_name]](session$input[[input_name]]),
      label = action_name
    )
  }
}

test_that("last returns the last element", {
  expect_equal(last(c(1, 2, 3)), 3)
  expect_equal(last(c("a", "b", "c")), "c")
  expect_equal(last(list(1, "a", TRUE)), TRUE)
  expect_equal(last(1), 1)
})

test_that("filter_null removes NULL elements", {
  expect_equal(filter_null(list(1, NULL, 3)), list(1, 3))
  expect_equal(filter_null(list(NULL, NULL)), list())
  expect_equal(filter_null(list(1, 2, 3)), list(1, 2, 3))
  expect_equal(filter_null(list()), list())
  expect_equal(filter_null(list(a = 1, b = NULL, c = 3)), list(a = 1, c = 3))
})

test_that("has_length checks if object has elements", {
  expect_true(has_length(c(1, 2, 3)))
  expect_true(has_length(1))
  expect_true(has_length("a"))
  expect_false(has_length(c()))
  expect_false(has_length(list()))
  expect_false(has_length(NULL))
})

library(testthat)

empty_board <- new_board()
non_empty_board <- new_board(
  blocks = list(
    new_dataset_block()
  )
)

test_that("dag extension ctor", {
  ext <- new_dag_extension()
  expect_s3_class(ext, "dag_extension")
})

test_that("ext_ui works", {
  ui <- dag_ext_ui("test", empty_board)
  expect_s3_class(ui, "shiny.tag.list")

  empty_state_container <- htmltools::tagQuery(ui)$find(
    ".dag-empty-state"
  )$selectedTags()[[1]]

  expect_null(empty_state_container$attribs$style)

  # Non empty board
  ui <- dag_ext_ui("test", non_empty_board)
  empty_state_container <- htmltools::tagQuery(ui)$find(
    ".dag-empty-state"
  )$selectedTags()[[1]]

  expect_identical(empty_state_container$attribs$style, "display: none;")

  # Custom dependencies
  deps <- htmltools::findDependencies(ui)
  dep_names <- chr_ply(deps, `[[`, "name")
  expect_contains(dep_names, c("rm-selection", "dag-empty-state"))
})


# Can we make the scope of blockr.dock before?
dock_view_proxy <- function(
  id,
  data = NULL,
  session = getDefaultReactiveDomain()
) {
  if (is.null(session)) {
    stop(
      "dock_view_proxy must be called from the server function of a Shiny app."
    )
  }
  structure(list(id = id, session = session), class = "dock_view_proxy")
}

test_board <- blockr.dock::new_dock_board(
  blocks = c(
    a = new_dataset_block("iris"),
    b = new_scatter_block(x = "Sepal.Length", y = "Sepal.Width"),
    c = new_head_block(),
    d = new_subset_block()
  ),
  links = list(from = "a", to = "b", input = "data"),
  stacks = c(
    stack_1 = blockr.dock::new_dock_stack(c("a", "b"), color = "#0000FF"),
    stack_2 = blockr.dock::new_dock_stack()
  ),
  extensions = new_dag_extension()
)

testServer(
  # Path to app.R is easier than trying to mock all board and dock stuff.
  #system.file("examples/stacks/app.R", package = "blockr.dag"),
  dag_ext_srv(NULL),
  args = list(
    board = reactiveValues(board = test_board),
    update = reactiveVal(NULL),
    # Mock dock returned value
    dock = list(
      layout = reactive(NULL), #blockr.dock::new_dock_layout()
      proxy = dock_view_proxy(
        "dock",
        session = MockShinySession$new()
      ),
      prev_active_group = reactiveVal(NULL)
    ),
    actions = blockr.dock::action_triggers(
      unlst(
        c(
          list(board_actions(test_board)),
          board_actions(new_dag_extension())
        )
      )
    )
  ),
  {
    # Test graph output
    session$output[[graph_id()]]
    expect_null(update())
    expect_s3_class(proxy, "g6_proxy")
    session$setInputs("graph-initialized" = TRUE)

    # Trigger draw link
    session$setInputs(
      added_edge = list(
        source = "a",
        target = "b",
        targetType = "node",
        sourcePort = "node-a-out",
        targetPort = "node-b-data"
      )
    )
    session$setInputs(
      added_edge = list(
        source = "a",
        target = "c",
        targetType = "node",
        sourcePort = "node-a-out",
        targetPort = "node-c-data"
      )
    )

    # Trigger append block
    session$setInputs(
      added_edge = list(
        source = "a",
        target = NULL,
        targetType = "canvas",
        sourcePort = "node-a-out",
        portType = "output"
      )
    )

    # Trigger prepend block
    session$setInputs(
      added_edge = list(
        source = "d",
        target = NULL,
        targetType = "canvas",
        sourcePort = "node-d-data",
        portType = "input"
      )
    )

    # Trigger brush_select
    session$setInputs(
      "graph-selected_node" = structure(
        c("a", "b"),
        eventType = "brush_select"
      )
    )

    # Trigger batch delete
    session$setInputs(
      "graph-batch_delete" = TRUE
    )

    # Remove node from combo
    stacks <- board_stacks(board$board)
    tmp_stack_1 <- stacks[[1]]
    stack_blocks(tmp_stack_1) <- stack_blocks(tmp_stack_1)[1]
    # Add node to combo
    tmp_stack_2 <- stacks[[2]]
    stack_blocks(tmp_stack_2) <- "d"
    mod_stacks <- as_stacks(set_names(
      list(tmp_stack_1, tmp_stack_2),
      board_stack_ids(board$board)
    ))

    # Update block title
    mod_blocks <- set_names(board_blocks(board$board)[c("a", "b")], c("a_new", "b_new"))

    update(
      list(
        blocks = list(
          add = as_blocks(new_dataset_block()),
          mod = mod_blocks
        ),
        stacks = list(
          add = as_stacks(blockr.dock::new_dock_stack("c", color = "#0000FF")),
          mod = mod_stacks,
          rm = c("stack_1")
        ),
        links = list(
          rm = "ab"
        )
      )
    )
    session$flushReact()
  }
)

test_that("extension_block_callback works", {
  ext_cb <- extension_block_callback(new_dag_extension())

  testServer(
    function(input, output, session) {
      conditions <- reactive({
        list(
          errors = input$errors
        )
      })
      res <- ext_cb(
        id = "test",
        board = test_board,
        update = reactiveVal(NULL),
        conditions = conditions,
        dag_extension = list(
          proxy = g6_proxy(
            "graph",
            session = session
          )
        ),
        session = session
      )
    },
    {
      expect_null(res)
      session$setInputs(
        `graph-initialized` = TRUE,
        errors = c("error1", "error2")
      )
      session$setInputs(errors = character(0))
    }
  )
})

blks_color <- function(blocks) {
  blockr.dock::blk_color(block_metadata(blocks)$category)
}

blks_icon <- function(blocks, size = 48L) {
  meta <- block_metadata(blocks)

  chr_mply(
    blockr.dock::blk_icon_data_uri,
    meta$icon,
    blockr.dock::blk_color(meta$category),
    MoreArgs = list(size = size)
  )
}

new_dummy_blk <- function(...) {
  new_data_block(
    function(id) {},
    function(id) tagList(),
    block_metadata = list(),
    class = "test_block",
    ...
  )
}

test_that("blks_color works", {
  blk1 <- new_dataset_block()
  cat1 <- blks_color(blk1)

  expect_type(cat1, "character")
  expect_length(cat1, 1L)

  # Unknown block
  default <- blockr.dock::blk_color(default_category())

  blk2 <- new_dummy_blk()
  cat2 <- blks_color(blk2)

  expect_identical(cat2, default)

  cat3 <- blks_color(c(blk1, blk2))

  expect_type(cat3, "character")
  expect_length(cat3, 2L)
})

test_that("blks_icon works", {
  blk1 <- new_dataset_block()
  icn1 <- blks_icon(blk1)

  expect_type(icn1, "character")
  expect_length(icn1, 1L)

  # Unknown block
  default <- blockr.dock::blk_icon_data_uri(
    default_icon(),
    blockr.dock::blk_color(default_category())
  )

  blk2 <- new_dummy_blk()
  icn2 <- blks_icon(new_dummy_blk())

  expect_identical(icn2, default)

  icn3 <- blks_icon(c(blk1, blk2))

  expect_type(icn3, "character")
  expect_length(icn3, 2L)
})

dag_ext_ui <- function(id, board) {
  ns <- NS(id)
  has_blocks <- length(board_blocks(board)) > 0

  # nolint start: line_length_linter
  icon_pointer <- paste0(
    '<svg width="20" height="20" viewBox="0 0 24 24" ',
    'fill="none" stroke="#9ca3af" stroke-width="2">',
    '<path d="M3 3l7.07 16.97 2.51-7.39 7.39-2.51L3 3z"/></svg>'
  )
  icon_plus <- paste0(
    '<svg width="16" height="16" viewBox="0 0 24 24" ',
    'fill="none" stroke="#d1d5db" stroke-width="2">',
    '<line x1="12" y1="5" x2="12" y2="19"/>',
    '<line x1="5" y1="12" x2="19" y2="12"/></svg>'
  )
  icon_workflow <- paste0(
    '<svg width="20" height="20" viewBox="0 0 24 24" ',
    'fill="none" stroke="#9ca3af" stroke-width="2">',
    '<rect x="3" y="3" width="6" height="6" rx="1"/>',
    '<rect x="15" y="3" width="6" height="6" rx="1"/>',
    '<rect x="9" y="15" width="6" height="6" rx="1"/>',
    '<path d="M6 9v3a3 3 0 0 0 3 3h6a3 3 0 0 0 3-3V9"/></svg>'
  )
  # nolint end

  tagList(
    tags$div(
      class = "dag-canvas-container",
      style = "position: relative; width: 100%; height: 100vh;",
      g6_output(graph_id(ns), height = "100vh"),
      tags$div(
        id = ns("empty-state"),
        class = "dag-empty-state",
        style = if (has_blocks) "display: none;" else NULL,
        tags$div(
          class = "dag-empty-state-content",
          tags$div(
            class = "dag-empty-state-icons",
            tags$div(class = "dag-empty-state-icon-box", HTML(icon_pointer)),
            HTML(icon_plus),
            tags$div(class = "dag-empty-state-icon-box", HTML(icon_workflow))
          ),
          tags$p(
            class = "dag-empty-state-title",
            "Start building your workflow"
          ),
          tags$div(
            class = "dag-empty-state-hints",
            tags$p(
              class = "dag-empty-state-hint",
              tags$kbd("Right-click"),
              " to add blocks"
            ),
            tags$p(
              class = "dag-empty-state-hint",
              tags$kbd("Shift"),
              " + ",
              tags$kbd("Drag"),
              " to connect"
            )
          )
        )
      )
    ),
    htmltools::htmlDependency(
      name = "rm-selection",
      version = pkg_version(),
      src = c(file = "assets"),
      script = file.path("js", "rm-sel.js"),
      package = pkg_name()
    ),
    htmltools::htmlDependency(
      name = "dag-empty-state",
      version = pkg_version(),
      src = c(file = "assets"),
      script = file.path("js", "empty-state.js"),
      stylesheet = file.path("css", "empty-state.css"),
      package = pkg_name()
    ),
    htmltools::htmlDependency(
      name = "copy-paste",
      version = pkg_version(),
      src = c(file = "assets"),
      script = file.path("js", "copy-paste.js"),
      package = pkg_name()
    )
  )
}

#' Context menu functions
#'
#' Functions for creating and working with context
#' menu entries.
#'
#' @param name Name of the context menu entry.
#' @param js JavaScript code to execute when the entry is selected.
#' @param action Action to perform when the entry is selected.
#' @param condition Condition to determine if the entry should be shown.
#' @param id Unique identifier for the context menu entry.
#'   Inferred from `name` if not provided
#' @param x Object to test or extract context menu items from.
#'
#' @details
#' \describe{
#'   \item{`new_context_menu_entry()`}{Creates a new context menu
#' entry with the specified name, JavaScript code, action function,
#' and display condition.}
#'   \item{`is_context_menu_entry()`}{
#' Tests whether an object is a valid context menu entry.}
#'   \item{`context_menu_items()`}{Generic function to
#' extract context menu items from various
#' objects like dock extensions, boards, or lists.}
#' }
#'
#' The `context_menu_items.dag_extension()` method
#' provides the following actions:
#' \itemize{
#'   \item Create link - Creates connections between workflow nodes.
#'   \item Remove block - Removes individual blocks from the workflow.
#'   \item Remove link - Removes connections between workflow nodes.
#'   \item Append block - Adds a new block after the selected node.
#'   \item Create stack - Creates a new workflow stack.
#'   \item Remove stack - Removes an entire workflow stack.
#'   \item Edit stack - Opens stack editing interface.
#'   \item Add block - Adds a new block to the canvas.
#' }
#'
#' @rdname ctx
#' @export
#' @return
#' \describe{
#'   \item{`new_context_menu_entry()`}{A context menu
#' entry object of class "context_menu_entry" containing
#' condition, action, and js functions, with name and id attributes.}
#'   \item{`is_context_menu_entry()`}{`TRUE` if `x` is
#' a context menu entry, `FALSE` otherwise.}
#'   \item{`context_menu_items()`}{A list of context
#' menu items for the given object.}
#' }
new_context_menu_entry <- function(
  name,
  js,
  action = NULL,
  condition = TRUE,
  id = tolower(gsub(" +", "_", name))
) {
  if (is.null(action)) {
    action <- function(...) NULL
  }

  if (isTRUE(condition)) {
    condition <- function(...) TRUE
  }

  if (is_string(js)) {
    js_string <- js
    js <- function(...) js_string
  }

  stopifnot(
    is.function(action),
    is.function(condition),
    is.function(js),
    is_string(id),
    is_string(name)
  )

  structure(
    list(condition = condition, action = action, js = js),
    name = name,
    id = id,
    class = "context_menu_entry"
  )
}

#' @rdname ctx
#' @export
is_context_menu_entry <- function(x) {
  inherits(x, "context_menu_entry")
}

context_menu_entry_id <- function(x) attr(x, "id")

context_menu_entry_name <- function(x) attr(x, "name")

context_menu_entry_condition <- function(x, ...) {
  x[["condition"]](...)
}

context_menu_entry_action <- function(x, actions, session = get_session()) {

  if (!is_context_menu_entry(x)) {

    res <- lapply(
      validate_context_menu_entries(x),
      context_menu_entry_action,
      actions,
      session
    )

    return(invisible(res))
  }

  stopifnot(is_context_menu_entry(x))

  fun <- x[["action"]]

  if (is.null(fun)) {
    return(invisible(NULL))
  }

  res <- fun(actions, session)

  if (!inherits(res, "Observer")) {
    blockr_abort(
      "Expecting context menu item server {context_menu_entry_id(x)} to ",
      "return an observer.",
      class = "context_menu_item_return_invalid"
    )
  }

  invisible(NULL)
}

context_menu_entry_js <- function(x, ns = NULL) {
  if (!is_context_menu_entry(x)) {
    validate_context_menu_entries(x)

    res <- paste(
      chr_ply(x, context_menu_entry_js, ns = ns),
      collapse = " else "
    )

    return(
      paste0("(value, target, current) => {\n", res, "\n}")
    )
  }

  if (is.null(ns)) {
    ns <- NS(NULL)
  }

  paste0(
    "if (value === '",
    context_menu_entry_id(x),
    "') {\n(",
    x[["js"]](ns),
    ")(value, target, current)\n}"
  )
}

build_context_menu <- function(x, ...) {
  if (!is_context_menu_entry(x)) {
    validate_context_menu_entries(x)

    res <- Filter(not_null, lapply(x, build_context_menu, ...))

    return(unname(res))
  }

  if (!context_menu_entry_condition(x, ...)) {
    return(NULL)
  }

  list(name = context_menu_entry_name(x), value = context_menu_entry_id(x))
}

validate_context_menu_entries <- function(x) {
  stopifnot(
    is.list(x),
    all(lgl_ply(x, is_context_menu_entry)),
    anyDuplicated(chr_ply(x, context_menu_entry_id)) == 0L
  )

  invisible(x)
}

#' @param x Object
#' @rdname ctx
#' @export
context_menu_items <- function(x) {
  UseMethod("context_menu_items")
}

#' @export
context_menu_items.dock_extension <- function(x) {
  list()
}

#' @export
context_menu_items.list <- function(x) {
  res <- lapply(x, context_menu_items)
  unlst(res)
}

#' @export
context_menu_items.dock_extensions <- function(x) {
  context_menu_items(as.list(x))
}

#' @export
context_menu_items.dock_board <- function(x) {
  context_menu_items(blockr.dock::dock_extensions(x))
}

test_that("context menu", {

  ext <- new_dag_extension()
  ctx <- context_menu_items(ext)

  node <- build_context_menu(ctx, target = list(type = "node"))

  expect_type(node, "list")
  expect_named(node, NULL)

  expect_setequal(
    chr_xtr(node, "value"),
    c("create_link", "remove_block", "append_block", "copy", "cut")
  )

  edge <- build_context_menu(ctx, target = list(type = "edge"))

  expect_type(edge, "list")
  expect_named(edge, NULL)

  expect_setequal(
    chr_xtr(edge, "value"),
    "remove_link"
  )

  canv <- build_context_menu(ctx, target = list(type = "canvas"))

  expect_type(canv, "list")
  expect_named(canv, NULL)

  expect_setequal(
    chr_xtr(canv, "value"),
    c("create_stack", "add_block", "paste")
  )

  comb <- build_context_menu(ctx, target = list(type = "combo"))

  expect_type(comb, "list")
  expect_named(comb, NULL)

  expect_setequal(
    chr_xtr(comb, "value"),
    c("remove_stack", "edit_stack", "copy", "cut")
  )
})

# --- Constructor ---

new_subboard <- function(
  blocks = blocks(),
  links = links(),
  stacks = stacks()
) {
  structure(
    list(blocks = blocks, links = links, stacks = stacks),
    class = "subboard"
  )
}

is_subboard <- function(x) {
  inherits(x, "subboard")
}

# --- Extraction helpers ---

subboard_blocks <- function(all_blocks, block_ids, stack_ids, all_stacks) {
  for (sid in intersect(stack_ids, names(all_stacks))) {
    block_ids <- union(block_ids, stack_blocks(all_stacks[[sid]]))
  }
  block_ids <- intersect(block_ids, names(all_blocks))
  if (length(block_ids) == 0L) {
    return(NULL)
  }
  all_blocks[block_ids]
}

subboard_links <- function(all_links, block_ids) {
  if (length(all_links) == 0L) {
    return(links())
  }
  keep <- all_links$from %in% block_ids & all_links$to %in% block_ids
  if (any(keep)) all_links[keep] else links()
}

subboard_stacks <- function(all_stacks, stack_ids, block_ids) {
  selected <- intersect(stack_ids, names(all_stacks))
  candidates <- setdiff(names(all_stacks), selected)
  inferred <- Filter(
    function(sid) {
      blks <- stack_blocks(all_stacks[[sid]])
      length(blks) > 0L && all(blks %in% block_ids)
    },
    candidates
  )
  all_stacks[union(selected, inferred)]
}

extract_subboard <- function(
  board,
  block_ids = character(),
  stack_ids = character()
) {
  stopifnot(
    is_board(board),
    is.character(block_ids),
    is.character(stack_ids)
  )

  all_blocks <- board_blocks(board)
  all_stacks <- board_stacks(board)

  blocks <- subboard_blocks(all_blocks, block_ids, stack_ids, all_stacks)
  if (is.null(blocks)) {
    return(NULL)
  }

  block_ids <- names(blocks)

  new_subboard(
    blocks = blocks,
    links = subboard_links(board_links(board), block_ids),
    stacks = subboard_stacks(all_stacks, stack_ids, block_ids)
  )
}

# --- S3 ser/deser methods ---

#' @exportS3Method blockr.core::blockr_ser subboard
blockr_ser.subboard <- function(x, blocks = NULL, ...) {
  list(
    object = class(x),
    payload = list(
      blocks = blockr_ser(x$blocks, blocks = blocks),
      links = blockr_ser(x$links),
      stacks = blockr_ser(x$stacks)
    )
  )
}

#' @exportS3Method blockr.core::blockr_deser subboard
blockr_deser.subboard <- function(x, data, ...) {
  new_subboard(
    blocks = blockr_deser(data$payload$blocks),
    links = blockr_deser(data$payload$links),
    stacks = blockr_deser(data$payload$stacks)
  )
}

# --- Live state capture ---

live_block_states <- function(board, block_ids) {
  stopifnot(is.character(block_ids))
  running <- board$blocks

  ids <- intersect(block_ids, names(running))
  lapply(
    lst_xtr(running[ids], "server", "state"),
    lapply,
    reval_if
  )
}

# --- ID remapping helpers ---

remap_blocks <- function(blocks, id_map) {
  blk_list <- as.list(blocks)
  names(blk_list) <- id_map[names(blk_list)]
  blk_list <- lapply(blk_list, function(blk) {
    block_name(blk) <- paste0(block_name(blk), " (copy)")
    blk
  })
  as_blocks(blk_list)
}

remap_links <- function(lnks, block_id_map, used_ids) {
  if (length(lnks) == 0L) {
    return(links())
  }
  remapped <- lapply(as.list(lnks), function(lnk) {
    new_link(
      from = unname(block_id_map[lnk$from]),
      to = unname(block_id_map[lnk$to]),
      input = lnk$input
    )
  })
  names(remapped) <- rand_names(old_names = used_ids, n = length(remapped))
  do.call(links, remapped)
}

remap_stacks <- function(stks, block_id_map, used_ids) {
  if (length(stks) == 0L) {
    return(stacks())
  }
  remapped <- lapply(as.list(stks), function(stk) {
    blockr.dock::new_dock_stack(
      blocks = unname(block_id_map[stack_blocks(stk)]),
      name = paste0(stack_name(stk), " (copy)"),
      color = blockr.dock::stack_color(stk)
    )
  })
  names(remapped) <- rand_names(old_names = used_ids, n = length(remapped))
  as_stacks(remapped)
}

remap_subboard_ids <- function(subboard, board) {
  stopifnot(
    is_subboard(subboard),
    is_board(board)
  )

  used_ids <- c(
    board_block_ids(board),
    board_link_ids(board),
    board_stack_ids(board)
  )

  new_block_ids <- rand_names(
    old_names = used_ids,
    n = length(subboard$blocks)
  )
  block_id_map <- set_names(new_block_ids, names(subboard$blocks))
  used_ids <- c(used_ids, new_block_ids)

  blocks <- remap_blocks(subboard$blocks, block_id_map)

  lnks <- remap_links(subboard$links, block_id_map, used_ids)
  used_ids <- c(used_ids, names(lnks))

  stks <- remap_stacks(subboard$stacks, block_id_map, used_ids)

  new_subboard(blocks, lnks, stks)
}

# --- Fixture ---

test_board <- new_board(
  blocks = c(
    a = new_dataset_block(),
    b = new_head_block(),
    c = new_dataset_block(),
    d = new_scatter_block()
  ),
  links = list(from = "a", to = "b", input = "data"),
  stacks = c(s1 = new_stack(c("a", "b")))
)

# --- new_subboard() ---

test_that("new_subboard() creates a subboard object", {
  sb <- new_subboard(blocks(), links(), stacks())
  expect_s3_class(sb, "subboard")
  expect_named(sb, c("blocks", "links", "stacks"))
  expect_length(sb$blocks, 0L)
  expect_length(sb$links, 0L)
  expect_length(sb$stacks, 0L)
})

# --- is_subboard() ---

test_that("is_subboard() identifies subboard objects", {
  sb <- new_subboard(blocks(), links(), stacks())
  expect_true(is_subboard(sb))
  expect_false(is_subboard(list()))
  expect_false(is_subboard(NULL))
})

# --- subboard_blocks() ---

test_that("subboard_blocks() returns explicit block_ids", {
  all_blocks <- board_blocks(test_board)
  all_stacks <- board_stacks(test_board)

  res <- subboard_blocks(all_blocks, c("a", "c"), character(), all_stacks)
  expect_length(res, 2L)
  expect_named(res, c("a", "c"))
})

test_that("subboard_blocks() resolves stack_ids to blocks", {
  all_blocks <- board_blocks(test_board)
  all_stacks <- board_stacks(test_board)

  res <- subboard_blocks(all_blocks, character(), "s1", all_stacks)
  expect_length(res, 2L)
  expect_true(all(c("a", "b") %in% names(res)))
})

test_that("subboard_blocks() combines block_ids and stack_ids", {
  all_blocks <- board_blocks(test_board)
  all_stacks <- board_stacks(test_board)

  res <- subboard_blocks(all_blocks, "c", "s1", all_stacks)
  expect_length(res, 3L)
  expect_true(all(c("a", "b", "c") %in% names(res)))
})

test_that("subboard_blocks() returns NULL when no match", {
  all_blocks <- board_blocks(test_board)
  all_stacks <- board_stacks(test_board)

  res <- subboard_blocks(all_blocks, "nonexistent", character(), all_stacks)
  expect_null(res)
})

test_that("subboard_blocks() filters out nonexistent IDs", {
  all_blocks <- board_blocks(test_board)
  all_stacks <- board_stacks(test_board)

  res <- subboard_blocks(all_blocks, c("a", "zzz"), character(), all_stacks)
  expect_length(res, 1L)
  expect_named(res, "a")
})

# --- subboard_links() ---

test_that("subboard_links() keeps link when both endpoints selected", {
  all_links <- board_links(test_board)
  res <- subboard_links(all_links, c("a", "b"))
  expect_length(res, 1L)
})

test_that("subboard_links() drops link when one endpoint missing", {
  all_links <- board_links(test_board)
  res <- subboard_links(all_links, "a")
  expect_length(res, 0L)
})

test_that("subboard_links() handles empty all_links", {
  res <- subboard_links(links(), c("a", "b"))
  expect_length(res, 0L)
})

# --- subboard_stacks() ---

test_that("subboard_stacks() returns explicit stack_ids", {
  all_stacks <- board_stacks(test_board)
  res <- subboard_stacks(all_stacks, "s1", character())
  expect_length(res, 1L)
  expect_true("s1" %in% names(res))
})

test_that("subboard_stacks() infers stacks when all member blocks selected", {
  all_stacks <- board_stacks(test_board)
  res <- subboard_stacks(all_stacks, character(), c("a", "b"))
  expect_length(res, 1L)
  expect_true("s1" %in% names(res))
})

test_that("subboard_stacks() does not infer stacks with partial coverage", {
  all_stacks <- board_stacks(test_board)
  res <- subboard_stacks(all_stacks, character(), "a")
  expect_length(res, 0L)
})

# --- extract_subboard() ---

test_that("extract_subboard() with blocks a,b returns subboard", {
  sb <- extract_subboard(test_board, block_ids = c("a", "b"))
  expect_s3_class(sb, "subboard")
  expect_length(sb$blocks, 2L)
  expect_length(sb$links, 1L)
  expect_length(sb$stacks, 1L)
})

test_that("extract_subboard() with stack_ids resolves blocks", {
  sb <- extract_subboard(test_board, stack_ids = "s1")
  expect_s3_class(sb, "subboard")
  expect_length(sb$blocks, 2L)
  expect_true(all(c("a", "b") %in% names(sb$blocks)))
  expect_length(sb$links, 1L)
  expect_length(sb$stacks, 1L)
})

test_that("extract_subboard() with isolated block returns no links/stacks", {
  sb <- extract_subboard(test_board, block_ids = "c")
  expect_s3_class(sb, "subboard")
  expect_length(sb$blocks, 1L)
  expect_length(sb$links, 0L)
  expect_length(sb$stacks, 0L)
})

test_that("extract_subboard() with no matching IDs returns NULL", {
  sb <- extract_subboard(test_board, block_ids = "zzz")
  expect_null(sb)
})

# --- blockr_ser / blockr_deser round-trip ---

test_that("blockr_ser produces expected structure", {
  sb <- extract_subboard(test_board, block_ids = c("a", "b"))
  ser <- blockr_ser(sb)

  expect_identical(ser$object, "subboard")
  expect_true("payload" %in% names(ser))
  expect_true(all(c("blocks", "links", "stacks") %in% names(ser$payload)))
})

test_that("blockr_ser / blockr_deser round-trip preserves structure", {
  sb <- extract_subboard(test_board, block_ids = c("a", "b"))
  ser <- blockr_ser(sb)
  deser <- blockr_deser(ser)

  expect_s3_class(deser, "subboard")
  expect_length(deser$blocks, length(sb$blocks))
  expect_length(deser$links, length(sb$links))
  expect_length(deser$stacks, length(sb$stacks))
})

test_that("full JSON pipeline round-trips", {
  sb <- extract_subboard(test_board, block_ids = c("a", "b"))
  json <- jsonlite::toJSON(blockr_ser(sb), auto_unbox = TRUE)
  data <- jsonlite::fromJSON(as.character(json), simplifyDataFrame = FALSE)
  deser <- blockr_deser(data)

  expect_s3_class(deser, "subboard")
  expect_length(deser$blocks, 2L)
  expect_length(deser$links, 1L)
  expect_length(deser$stacks, 1L)
})

test_that("empty subboard round-trips", {
  sb <- new_subboard(blocks(), links(), stacks())
  ser <- blockr_ser(sb)
  deser <- blockr_deser(ser)

  expect_s3_class(deser, "subboard")
  expect_length(deser$blocks, 0L)
  expect_length(deser$links, 0L)
  expect_length(deser$stacks, 0L)
})

# --- remap_blocks() ---

test_that("remap_blocks() remaps IDs and adds (copy) suffix", {
  sb <- extract_subboard(test_board, block_ids = c("a", "b"))
  id_map <- c(a = "x1", b = "x2")

  remapped <- remap_blocks(sb$blocks, id_map)
  expect_named(remapped, c("x1", "x2"))
  expect_true(all(grepl(
    "\\(copy\\)$",
    vapply(
      as.list(remapped),
      block_name,
      character(1)
    )
  )))
})

# --- remap_links() ---

test_that("remap_links() remaps from/to and generates fresh IDs", {
  sb <- extract_subboard(test_board, block_ids = c("a", "b"))
  block_id_map <- c(a = "x1", b = "x2")
  used_ids <- c("a", "b", "x1", "x2", names(sb$links))

  remapped <- remap_links(sb$links, block_id_map, used_ids)
  expect_length(remapped, length(sb$links))

  lnk <- as.list(remapped)[[1]]
  expect_identical(lnk$from, "x1")
  expect_identical(lnk$to, "x2")

  expect_true(!any(names(remapped) %in% used_ids))
})

test_that("remap_links() handles empty links", {
  res <- remap_links(links(), c(a = "x1"), c("a", "x1"))
  expect_length(res, 0L)
})

# --- remap_stacks() ---

test_that("remap_stacks() remaps block refs and adds (copy) suffix", {
  sb <- extract_subboard(test_board, block_ids = c("a", "b"))
  block_id_map <- c(a = "x1", b = "x2")
  used_ids <- c("a", "b", "x1", "x2", "s1")

  remapped <- remap_stacks(sb$stacks, block_id_map, used_ids)
  expect_length(remapped, 1L)

  stk <- as.list(remapped)[[1]]
  expect_true(all(stack_blocks(stk) %in% c("x1", "x2")))
  expect_true(grepl("\\(copy\\)$", stack_name(stk)))

  expect_true(!any(names(remapped) %in% used_ids))
})

# --- remap_subboard_ids() ---

test_that("remap_subboard_ids() produces no collisions", {
  sb <- extract_subboard(test_board, block_ids = c("a", "b"))
  remapped <- remap_subboard_ids(sb, test_board)

  orig_ids <- c(
    board_block_ids(test_board),
    board_link_ids(test_board),
    board_stack_ids(test_board)
  )

  expect_true(!any(names(remapped$blocks) %in% orig_ids))
  expect_true(!any(names(remapped$links) %in% orig_ids))
  expect_true(!any(names(remapped$stacks) %in% orig_ids))
})

test_that("remap_subboard_ids() preserves counts", {
  sb <- extract_subboard(test_board, block_ids = c("a", "b"))
  remapped <- remap_subboard_ids(sb, test_board)

  expect_length(remapped$blocks, length(sb$blocks))
  expect_length(remapped$links, length(sb$links))
  expect_length(remapped$stacks, length(sb$stacks))
})

test_that("remap_subboard_ids() adds (copy) suffix to block names", {
  sb <- extract_subboard(test_board, block_ids = c("a", "b"))
  remapped <- remap_subboard_ids(sb, test_board)

  nms <- vapply(as.list(remapped$blocks), block_name, character(1))
  expect_true(all(grepl("\\(copy\\)$", nms)))
})

test_that("remap_subboard_ids() links reference new block IDs", {
  sb <- extract_subboard(test_board, block_ids = c("a", "b"))
  remapped <- remap_subboard_ids(sb, test_board)

  new_block_ids <- names(remapped$blocks)
  lnk <- as.list(remapped$links)[[1]]
  expect_true(lnk$from %in% new_block_ids)
  expect_true(lnk$to %in% new_block_ids)
})

test_that("remap_subboard_ids() stack block refs are subset of new block IDs", {
  sb <- extract_subboard(test_board, block_ids = c("a", "b"))
  remapped <- remap_subboard_ids(sb, test_board)

  new_block_ids <- names(remapped$blocks)
  stk <- as.list(remapped$stacks)[[1]]
  expect_true(all(stack_blocks(stk) %in% new_block_ids))
})

dag_ext_srv <- function(graph) {
  function(id, board, update, dock, actions, ...) {
    dot_args <- list(...)

    moduleServer(
      id,
      function(input, output, session) {
        initial_board <- isolate(board$board)

        context_menu <- context_menu_items(initial_board)

        ctx_path <- session$registerDataObj(
          name = "context-menu-items",
          data = list(),
          filterFunc = function(data, req) {
            body_bytes <- req$rook.input$read(-1)
            res <- jsonlite::toJSON(
              build_context_menu(
                context_menu,
                board = initial_board,
                target = jsonlite::fromJSON(rawToChar(body_bytes))
              )
            )
            httpResponse(
              content_type = "application/json",
              content = res
            )
          }
        )

        toolbar <- toolbar_items(initial_board)

        init_g6(
          board = initial_board,
          graph = graph,
          path = ctx_path,
          ctx = context_menu,
          tools = toolbar,
          session = session
        )

        proxy <- blockr_g6_proxy(session)

        context_menu_entry_action(context_menu, actions, session)
        toolbar_item_action(toolbar, actions, session)

        setup_remove_elements_kbd()
        setup_copy_paste_kbd()

        actions_observers(actions, proxy)

        update_observer(update, board, proxy)

        observeEvent(
          input[[paste0(graph_id(), "-selected_node")]],
          {
            sel <- input[[paste0(graph_id(), "-selected_node")]]
            evt <- attr(sel, "eventType")

            if (length(sel) == 1L && !identical(evt, "brush_select")) {
              blockr.dock::show_panel(
                from_g6_node_id(sel),
                board$board,
                dock
              )
            }
          },
          label = "selected_node"
        )

        empty_state_observer(board, session)

        list(
          state = list(
            graph = reactive(
              input[[paste0(graph_id(), "-state")]],
              label = "graph_state"
            )
          ),
          proxy = proxy
        )
      }
    )
  }
}

update_observer <- function(update, board, proxy) {
  observeEvent(
    update(),
    {
      upd <- update()

      if (length(upd$blocks$add)) {
        add_nodes(upd$blocks$add, board$board, proxy)
      }

      if (length(upd$blocks$mod)) {
        update_nodes(upd$blocks$mod, board$board, proxy)
      }

      if (length(upd$links$add)) {
        blocks <- board_blocks(board$board)
        # If new blocks are added in the same update as new links, for instance
        # during an append action, we need to consider the new blocks that are in
        # upd$blocks$add and not yet in the board. Otherwise, we can possibly end up
        # with non existing target/source nodes.
        if (length(upd$blocks$add)) {
          blocks <- c(blocks, upd$blocks$add)
        }
        add_edges(upd$links$add, blocks, proxy)
      }

      if (length(upd$stacks$add)) {
        add_combos(upd$stacks$add, proxy)
      }

      if (length(upd$stacks$mod)) {
        update_combos(upd$stacks$mod, board$board, proxy)
      }

      if (length(upd$stacks$rm)) {
        remove_combos(upd$stacks$rm, proxy)
      }

      if (length(upd$links$rm)) {
        remove_edges(upd$links$rm, proxy = proxy)
      }

      if (length(upd$blocks$rm)) {
        remove_nodes(upd$blocks$rm, proxy = proxy)
      }
    },
    label = "update_observer"
  )
}

actions_observers <- function(actions, proxy) {
  input <- proxy$session$input

  observeEvent(
    input[[paste0(graph_id(), "-batch_delete")]],
    actions[["remove_selected_action"]](
      input[[paste0(graph_id(), "-batch_delete")]]
    ),
    label = "batch_delete"
  )

  observeEvent(
    req(input$added_edge$targetType != "canvas"),
    {
      actions[["draw_link_action"]](input$added_edge)
    },
    label = "draw_link"
  )

  observeEvent(
    input[[paste0(graph_id(), "-copy_selected")]],
    actions[["copy_selected_action"]](
      input[[paste0(graph_id(), "-copy_selected")]]
    ),
    label = "copy_selected"
  )

  observeEvent(
    input[[paste0(graph_id(), "-cut_selected")]],
    actions[["cut_selected_action"]](
      input[[paste0(graph_id(), "-cut_selected")]]
    ),
    label = "cut_selected"
  )

  observeEvent(
    input[[paste0(graph_id(), "-paste_clipboard")]],
    actions[["paste_action"]](
      input[[paste0(graph_id(), "-paste_clipboard")]]
    ),
    label = "paste_clipboard"
  )

  # Append/prepend from canvas drop
  observeEvent(
    req(input$added_edge$targetType == "canvas"),
    {
      edge <- input$added_edge
      req(edge$portType)

      switch(
        edge$portType,
        output = actions[["append_block_action"]](edge$source),
        input = actions[["prepend_block_action"]](edge$source)
      )
    },
    label = "canvas_drop"
  )

  # Append/prepend on port click: FIXME -> disabled due to critical issue
  # with input ports which cannot be clicked :)
  # observeEvent(
  #   input[[paste0(graph_id(), "-selected_port")]],
  #   {
  #     el <- input[[paste0(graph_id(), "-selected_port")]]
  #     req(el$type)

  #     switch(
  #       el$type,
  #       output = actions[["append_block_action"]](from_g6_node_id(el$node)),
  #       input = actions[["prepend_block_action"]](from_g6_node_id(el$node))
  #     )
  #   }
  # )
}

empty_state_observer <- function(board, session) {
  ns <- session$ns

  observeEvent(
    board$board,
    {
      has_blocks <- length(board_blocks(board$board)) > 0
      session$sendCustomMessage(
        "update-empty-state",
        list(
          id = ns("empty-state"),
          show = !has_blocks
        )
      )
    },
    label = "empty_state"
  )
}

to_g6_node_id <- function(x) {
  if (length(x)) {
    x <- paste0("node-", x)
  }

  x
}

from_g6_node_id <- function(x) {
  sub("^node-", "", x)
}

to_g6_edge_id <- function(x) {
  if (length(x)) {
    x <- paste0("edge-", x)
  }

  x
}

from_g6_edge_id <- function(x) {
  sub("^edge-", "", x)
}

to_g6_combo_id <- function(x) {
  if (length(x)) {
    x <- paste0("combo-", x)
  }

  x
}

from_g6_combo_id <- function(x) {
  sub("^combo-", "", x)
}

from_g6_port_id <- function(x, node) {
  sub(paste0("^", node, "-"), "", x)
}

to_g6_port_id <- function(x, node) {
  if (length(x)) {
    x <- paste0(node, "-", x)
  }
  x
}

g6_from_board <- function(board) {
  stopifnot(is_board(board))

  graph <- g6_data_from_board(board)

  g6(
    nodes = graph_nodes(graph),
    edges = graph_edges(graph),
    combos = graph_combos(graph)
  )
}

g6_from_graph <- function(graph) {
  stopifnot(is_graph(graph))

  g6(
    nodes = graph_nodes(graph),
    edges = graph_edges(graph),
    combos = graph_combos(graph)
  )
}

set_g6_options <- function(graph, ...) {
  g6_options(
    graph,
    ...,
    # Required so that elements are in the DOM
    # for shinytest2 e2e
    renderer = JS("() => new SVGRenderer()"),
    animation = FALSE,
    node = list(
      type = "custom-image-node",
      style = list(
        labelFill = "#6b7280",
        labelBackground = TRUE,
        labelBackgroundFill = "#f3f4f6",
        labelBackgroundStroke = "#e5e7eb",
        labelBackgroundRadius = 4,
        labelPlacement = "bottom",
        labelOffsetY = 8,
        labelBackgroundLineWidth = 1,
        labelBackgroundRadius = 4,
        labelBackgroundOpacity = 1,
        labelPadding = c(1, 6, 1, 6),
        labelFontSize = 11,
        labelFontFamily = "Open Sans, system-ui, sans-serif"
      ),
      state = list(
        selected = list(
          labelBackgroundFill = "#dbeafe",
          labelBackgroundStroke = "#0D99FF",
          labelFontWeight = 700
        )
      )
    ),
    combo = list(
      animation = FALSE,
      badge = TRUE,
      type = "rect-combo-with-extra-button",
      style = list(
        # more bottom padding, because of the badge
        padding = c(20, 20, 40, 20)
      )
    ),
    edge = list(
      type = "cubic-vertical",
      style = list(
        zIndex = -1,
        endArrow = TRUE,
        stroke = "#D1D5DB",
        lineWidth = 2,
        increasedLineWidthForHitTesting = 25,
        curveOffset = list(15, -15),
        labelOpacity = 0,
        labelFill = "#6b7280",
        labelBackground = TRUE,
        labelBackgroundFill = "#FFFFFF",
        labelBackgroundRadius = 4,
        labelBackgroundOpacity = 0,
        labelPadding = c(2, 4)
      ),
      state = list(
        active = list(
          stroke = "#D1D5DB",
          lineWidth = 2,
          shadowBlur = 0,
          shadowColor = "transparent",
          halo = FALSE,
          labelOpacity = 1,
          labelBackgroundOpacity = 1
        ),
        selected = list(
          stroke = "#9ca3af",
          lineWidth = 2,
          shadowBlur = 4,
          shadowColor = "rgba(156, 163, 175, 0.5)",
          halo = FALSE
        )
      )
    )
  )
}

set_g6_layout <- function(graph) {
  g6_layout(
    graph,
    layout = antv_dagre_layout(
      begin = c(150, 150),
      nodesep = 50,
      ranksep = 50,
      sortByCombo = TRUE
    )
  )
}

set_g6_behaviors <- function(graph, ..., ns) {
  g6_behaviors(
    graph,
    ...,
    zoom_canvas(),
    drag_canvas(
      enable = JS(
        "(e) => {
          return e.targetType === 'canvas' && !e.shiftKey && !e.altKey;
        }"
      )
    ),
    # So we can add node to stack from the UI by drag and drop
    # Disable drag when edge creation from port is active
    drag_element(
      enable = JS(
        "(e) => {
          if (e.shiftKey || e.altKey) return false;
          // Access graph via HTMLWidgets and check if edge creation is in progress
          const target = e.nativeEvent?.target;
          const graph = HTMLWidgets.find(`#${target?.closest?.('.g6')?.id}`)?.getWidget();
          try {
            if (graph?.getNodeData?.('g6-create-edge-assist-node-id')) return false;
          } catch (err) {}
          return true;
        }"
      ),
      # For now, we prevent nodes from being dropped outside combo.
      dropEffect = "move"
    ),
    click_select(multiple = TRUE),
    brush_select(
      # Option key on mac
      trigger = "Alt",
      enableElements = c("node", "edge", "combo"),
      immediately = TRUE,
      outputId = graph_id(ns)
    ),
    collapse_expand(),
    # Show edge labels on hover (CSS handles the transition animation)
    hover_activate(
      animation = FALSE,
      enable = JS("(e) => e.targetType === 'edge'")
    ),
    # avoid conflict with internal function
    g6R::create_edge(
      enable = JS(
        "(e) => {
          return e.targetType === 'node' && e.targetType !== 'combo'
        }"
      ),
      onFinish = JS(
        sprintf(
          "(edge) => {
            const graph = HTMLWidgets.find('#%s').getWidget();
            // For canvas drops, the assist node is already removed, so check targetType first
            if (edge.targetType === 'canvas') {
              Shiny.setInputValue(
                '%s',
                {
                  id: edge.id,
                  source: edge.source.replace(/^node-/, ''),
                  target: null,
                  targetType: 'canvas',
                  sourcePort: edge.style?.sourcePort,
                  portType: edge.style?.portType
                }
              );
              return;
            }
            const targetType = graph.getElementType(edge.target);
            // Avoid to create edges in combos. If so, we remove it
            if (targetType === 'combo') {
              graph.removeEdgeData([edge.id]);
            } else {
              Shiny.setInputValue(
                '%s',
                {
                  id: edge.id,
                  source: edge.source.replace(/^node-/, ''),
                  target: edge.target.replace(/^node-/, ''),
                  targetType: edge.targetType,
                  sourcePort: edge.style.sourcePort,
                  targetPort: edge.style.targetPort
                }
              );
              // We will recreate the edge from the R side with correct ID
              graph.removeEdgeData([edge.id]);
            }
          }",
          graph_id(ns),
          ns("added_edge"),
          ns("added_edge")
        )
      )
    )
  )
}

set_g6_plugins <- function(graph, ..., ns, path, ctx, tools) {
  g6_plugins(
    graph,
    ...,
    context_menu(
      enable = JS(
        "(e) => {
          let cond = e.targetType === 'edge' ||
            e.targetType === 'node' ||
            e.targetType === 'canvas' ||
            e.targetType === 'combo';
          return cond;
        }"
      ),
      onClick = JS(context_menu_entry_js(ctx, ns)),
      getItems = JS(
        sprintf(
          "async (e) => {
            var body;
            if (e.targetType === 'canvas') {
              body = {type: e.targetType};
            } else {
              body = {
                id: e.target.id.replace(/^(node|edge|combo)-/, ''),
                type: e.targetType
              };
            }
            const response = await fetch(
              '%s',
              {
                method: 'POST',
                headers: {
                  'Accept': 'application/json',
                  'Content-Type': 'application/json'
                },
                body: JSON.stringify(body)
              }
            );
            const items = await response.json();
            return items;
          }",
          path
        )
      )
    ),
    toolbar(
      style = list(
        backgroundColor = "#f5f5f5",
        padding = "8px",
        boxShadow = "0 2px 8px rgba(0, 0, 0, 0.15)",
        borderRadius = "8px",
        border = "1px solid #e8e8e8",
        opacity = "0.9",
        marginTop = "12px",
        marginLeft = "12px"
      ),
      position = "left",
      getItems = JS(build_toolbar(tools)),
      onClick = JS(toolbar_item_js(tools, ns))
    )
  )
}

blockr_g6_proxy <- function(session = get_session()) {
  g6_proxy(graph_id(session$ns), session = session)
}

init_g6 <- function(board, graph = NULL, ..., session = get_session()) {
  ns <- session$ns

  if (is.null(graph)) {
    res <- g6_from_board(board)
  } else {
    res <- tryCatch(
      g6_from_graph(as_graph(graph)),
      error = function(e) {
        showNotification(
          sprintf(
            "Failed to initialize graph from provided graph object: %s.
            \nFalling back to graph generated from board.",
            conditionMessage(e)
          ),
          type = "error"
        )
        g6_from_board(board)
      }
    )
  }

  res <- set_g6_options(res)
  res <- set_g6_layout(res)
  res <- set_g6_behaviors(res, ns = ns)
  res <- set_g6_plugins(res, ..., ns = ns)

  # For shinytest2 e2e
  preprocess_graph_state()
  preprocess_mouse_position()

  session$output[[graph_id()]] <- render_g6(res)

  invisible(res)
}

#' Check if a block is variadic
#'
#' A block is variadic if it has NA arity and no named inputs.
#'
#' @param block A block object
#' @return Logical indicating if the block is variadic
#' @keywords internal
is_variadic_block <- function(block) {
  is_block(block) &&
    is.na(blockr.core::block_arity(block)) &&
    length(blockr.core::block_inputs(block)) == 0
}

#' Resolve target port IDs for links
#'
#' @param links Board links
#' @param blocks Board blocks
#' @return Character vector of target port IDs
#' @keywords internal
resolve_target_ports <- function(links, blocks) {
  if (length(links) == 0) {
    return()
  }

  target_id <- to_g6_node_id(links$to)

  vapply(
    seq_along(links$to),
    function(i) {
      blk <- blocks[[links$to[i]]]
      if (is_variadic_block(blk)) {
        paste0(target_id[i], "-in")
      } else {
        paste0(target_id[i], "-", links$input[i])
      }
    },
    character(1)
  )
}

#' @rdname g6r
#' @param links Board links.
#' @param blocks Board blocks
g6_edges_from_links <- function(links, blocks) {
  if (length(links) == 0) {
    return()
  }
  source_id <- to_g6_node_id(links$from)
  target_ports <- resolve_target_ports(links, blocks)

  res <- map(
    g6_edge,
    id = to_g6_edge_id(names(links)),
    source = source_id,
    target = to_g6_node_id(links$to),
    style = map(
      list,
      labelText = links$input,
      # Currently all nodes only have 1 output that is called out
      sourcePort = paste0(source_id, "-out"),
      # Note: targetPort label is built in create_block_ports()
      # from the link input name so if we prefix by the
      # node id, we are good to go!
      targetPort = target_ports
    ),
    MoreArgs = list(type = "cubic-vertical")
  )

  if (length(res)) {
    do.call(g6_edges, res)
  } else {
    res
  }
}

#' Create block ports for g6 node
#' @param block Block object.
#' @param id Block ID.
#' @keywords internal
create_block_ports <- function(block, id) {
  inputs <- blockr.core::block_inputs(block)
  arity <- blockr.core::block_arity(block)
  input_ports <- list()
  fill_col <- blks_color(block)

  # variadic input
  if (is_variadic_block(block)) {
    input_ports <- list(g6_input_port(
      key = sprintf("%s-in", id),
      label = "in",
      arity = Inf,
      visibility = "hover",
      placement = "top",
      fill = fill_col
    ))
  } else if (length(inputs) == 1 && arity == 1) {
    # Mono input
    input_ports <- list(g6_input_port(
      key = sprintf("%s-%s", id, inputs[1]),
      label = inputs[1],
      arity = 1,
      visibility = "hover",
      placement = "top",
      fill = fill_col
    ))
  } else if (length(inputs) > 1) {
    # Multi input
    n <- length(inputs)
    if (n == 1) {
      xs <- 0.5
    } else {
      xs <- seq(0.15, 0.85, length.out = n)
    }
    input_ports <- lapply(seq_along(inputs), function(i) {
      g6_input_port(
        key = sprintf("%s-%s", id, inputs[i]),
        label = inputs[i],
        arity = 1,
        visibility = "hover",
        placement = c(xs[i], 0),
        fill = fill_col
      )
    })
  }

  out_id <- sprintf("%s-out", id)

  ports <- c(
    input_ports,
    list(g6_output_port(
      key = out_id,
      label = NULL,
      arity = Inf,
      visibility = "hover",
      placement = "bottom",
      fill = fill_col
    ))
  )
  do.call(g6_ports, ports)
}

#' Get children relationships from links
#'
#' @param links Board links
#' @return Named list where each element is a list of child node IDs (for JSON array conversion)
#' @keywords internal
get_children_from_links <- function(links) {
  if (length(links) == 0) {
    return(list())
  }

  # Get unique source nodes
  sources <- unique(links$from)

  # For each source, get all its target nodes as a list (for JSON array conversion)
  children <- lapply(sources, function(src) {
    targets <- links$to[links$from == src]
    as.list(to_g6_node_id(unique(targets)))
  })

  set_names(children, to_g6_node_id(sources))
}

#' @rdname g6r
#' @param blocks Board blocks.
#' @param stacks Board stacks.
#' @param children Named list of children node IDs (optional).
#' @keywords internal
g6_nodes_from_blocks <- function(blocks, stacks, children = NULL) {
  stk_blks <- lapply(stacks, stack_blocks)

  stk_blks <- set_names(
    as.list(rep(names(stk_blks), lengths(stk_blks))),
    do.call("c", stk_blks)
  )

  ids <- to_g6_node_id(names(blocks))

  # Build base arguments for g6_node
  base_args <- list(
    id = ids,
    style = map(
      list,
      src = blks_icon(blocks, size = 48),
      labelText = chr_ply(blocks, block_name),
      MoreArgs = list(size = 48)
    ),
    combo = lapply(stk_blks[names(blocks)], to_g6_combo_id),
    ports = map(create_block_ports, blocks, ids),
    collapse = lapply(blocks, function(block) {
      g6_collapse_options(visibility = "hover", stroke = "#D1D5DB")
    })
  )

  # Only add children if we have them and they're non-empty
  if (!is.null(children) && length(children) > 0) {
    node_children <- lapply(ids, function(id) {
      child_ids <- children[[id]]
      # Return NULL for nodes without children (will be removed by dropNulls in g6_node)
      if (is.null(child_ids) || length(child_ids) == 0) {
        NULL
      } else {
        child_ids
      }
    })
    base_args$children <- node_children
  }

  res <- do.call(map, c(list(g6_node), base_args))

  if (length(res)) {
    do.call(g6_nodes, res)
  } else {
    res
  }
}

#' @rdname g6r
#' @param stacks Board stacks.
#' @keywords internal
g6_combos_data_from_stacks <- function(stacks) {
  colors <- blockr.dock::stack_color(stacks)

  colors[is.na(colors)] <- blockr.dock::suggest_new_colors(
    coal(unlst(colors[!is.na(colors)]), character()),
    sum(is.na(colors))
  )

  res <- map(
    g6_combo,
    id = to_g6_combo_id(names(stacks)),
    style = map(
      list,
      fill = colors,
      shadowColor = colors,
      collapsedFill = colors,
      iconFill = colors,
      labelText = chr_ply(stacks, stack_name),
      MoreArgs = list(
        fillOpacity = 0.1,
        labelPlacement = "top",
        lineWidth = 0,
        radius = 8
      )
    ),
    collapse = lapply(stacks, function(stack) {
      g6_collapse_options(visibility = "hover", stroke = "#D1D5DB")
    })
  )

  if (length(res)) {
    do.call(g6_combos, res)
  } else {
    res
  }
}

#' Create network data from board
#'
#' @keywords internal
#' @param board Board object.
#' @rdname g6r
g6_data_from_board <- function(board) {
  # Cold start
  links <- board_links(board)
  blocks <- board_blocks(board)
  stacks <- board_stacks(board)

  # Get children relationships from links
  children <- get_children_from_links(links)

  edges_data <- g6_edges_from_links(links, blocks)
  combos_data <- g6_combos_data_from_stacks(stacks)
  nodes_data <- g6_nodes_from_blocks(blocks, stacks, children)

  new_graph(
    nodes = nodes_data,
    edges = edges_data,
    combos = combos_data
  )
}

remove_nodes <- function(nodes, asis = FALSE, proxy = blockr_g6_proxy()) {
  if (!isTRUE(asis)) {
    nodes <- to_g6_node_id(nodes)
  }

  g6_remove_nodes(proxy, nodes)

  invisible()
}

remove_edges <- function(edges, asis = FALSE, proxy = blockr_g6_proxy()) {
  if (!isTRUE(asis)) {
    edges <- to_g6_edge_id(edges)
  }

  g6_remove_edges(proxy, edges)

  invisible()
}

remove_combos <- function(combos, asis = FALSE, proxy = blockr_g6_proxy()) {
  if (!isTRUE(asis)) {
    combos <- to_g6_combo_id(combos)
  }

  g6_remove_combos(proxy, combos)

  invisible()
}

add_nodes <- function(blocks, board, proxy = blockr_g6_proxy()) {
  nodes <- g6_nodes_from_blocks(blocks, board_stacks(board))

  mouse_pos <- proxy$session$input[[paste0(graph_id(), "-mouse_position")]]
  base_x <- mouse_pos$x %||% 150
  base_y <- mouse_pos$y %||% 150

  if (length(nodes) == 1) {
    nodes[[1]]$style$x <- base_x
    nodes[[1]]$style$y <- base_y
  } else if (length(nodes) > 1) {
    for (i in seq_along(nodes)) {
      nodes[[i]]$style$x <- base_x
      nodes[[i]]$style$y <- base_y + (i - 1) * 130
    }
  }

  g6_add_nodes(proxy, nodes)
  invisible()
}

update_nodes <- function(blocks, board, proxy = blockr_g6_proxy()) {
  nodes <- g6_nodes_from_blocks(blocks, board_stacks(board))
  g6_update_nodes(proxy, nodes)

  invisible()
}

add_edges <- function(links, blocks, proxy = blockr_g6_proxy()) {
  edges <- g6_edges_from_links(links, blocks)
  g6_add_edges(proxy, edges)

  invisible()
}

add_combos <- function(stacks, proxy = blockr_g6_proxy()) {
  combos <- g6_combos_data_from_stacks(stacks)
  g6_add_combos(proxy, combos)

  # Add nodes to stacks if any
  add_nodes_to_combos(stacks, proxy)

  invisible()
}

add_nodes_to_combos <- function(stacks, proxy = blockr_g6_proxy()) {
  map(
    add_nodes_to_combo,
    lapply(stacks, stack_blocks),
    names(stacks),
    MoreArgs = list(proxy = proxy)
  )

  invisible()
}

add_nodes_to_combo <- function(block_ids, stack_id, proxy = blockr_g6_proxy()) {
  g6_update_nodes(
    proxy,
    map(
      list,
      id = to_g6_node_id(block_ids),
      MoreArgs = list(combo = to_g6_combo_id(stack_id))
    )
  )

  invisible()
}

remove_nodes_from_combo <- function(block_ids, proxy = blockr_g6_proxy()) {
  g6_update_nodes(
    proxy,
    map(list, id = to_g6_node_id(block_ids), MoreArgs = list(combo = NULL))
  )

  invisible()
}

update_combos <- function(stacks, board, proxy = blockr_g6_proxy()) {
  g6_update_combos(proxy, g6_combos_data_from_stacks(stacks))

  cur_stacks <- board_stacks(board)

  for (id in names(stacks)) {
    cur_stack_blocks <- stack_blocks(cur_stacks[[id]])
    new_stack_blocks <- stack_blocks(stacks[[id]])

    to_add <- setdiff(new_stack_blocks, cur_stack_blocks)
    to_remove <- setdiff(cur_stack_blocks, new_stack_blocks)

    # Prevent no-oops from flooding the JS handlers
    if (length(to_remove)) {
      remove_nodes_from_combo(to_remove, proxy)
    }
    if (length(to_add)) {
      add_nodes_to_combo(to_add, id, proxy)
    }
  }

  invisible()
}

setup_remove_elements_kbd <- function(
  key = "Backspace",
  session = get_session()
) {
  input <- session$input
  ns <- session$ns
  observeEvent(
    req(input[[paste0(graph_id(), "-initialized")]]),
    {
      session$sendCustomMessage(
        "setup-remove-selected-elements",
        # TBD: key can be a board option
        list(
          key = key,
          id = graph_id(ns)
        )
      )
    },
    once = TRUE,
    label = "setup_remove_kbd"
  )
}

setup_copy_paste_kbd <- function(session = get_session()) {
  input <- session$input
  ns <- session$ns
  observeEvent(
    req(input[[paste0(graph_id(), "-initialized")]]),
    {
      session$sendCustomMessage(
        "setup-copy-paste",
        list(id = graph_id(ns))
      )
    },
    once = TRUE,
    label = "setup_copy_paste_kbd"
  )
}

library(testthat)

test_that("to_g6_node_id works correctly", {
  expect_equal(to_g6_node_id("test1"), "node-test1")
  expect_equal(
    to_g6_node_id(c("test1", "test2")),
    c("node-test1", "node-test2")
  )
  expect_equal(to_g6_node_id(character(0)), character(0))
  expect_equal(to_g6_node_id(NULL), NULL)
  expect_equal(to_g6_node_id(""), "node-")
  expect_equal(to_g6_node_id(123), "node-123")
})

test_that("from_g6_node_id works correctly", {
  expect_equal(from_g6_node_id("node-test1"), "test1")
  expect_equal(
    from_g6_node_id(c("node-test1", "node-test2")),
    c("test1", "test2")
  )
  expect_equal(from_g6_node_id("test1"), "test1")
  expect_equal(from_g6_node_id(""), "")
  expect_equal(from_g6_node_id("node-"), "")
  expect_equal(from_g6_node_id("node-node-test"), "node-test")
})

test_that("to_g6_edge_id works correctly", {
  expect_equal(to_g6_edge_id("edge1"), "edge-edge1")
  expect_equal(
    to_g6_edge_id(c("edge1", "edge2")),
    c("edge-edge1", "edge-edge2")
  )
  expect_equal(to_g6_edge_id(character(0)), character(0))
  expect_equal(to_g6_edge_id(NULL), NULL)
  expect_equal(to_g6_edge_id(""), "edge-")
  expect_equal(to_g6_edge_id(456), "edge-456")
})

test_that("from_g6_edge_id works correctly", {
  expect_equal(from_g6_edge_id("edge-edge1"), "edge1")
  expect_equal(
    from_g6_edge_id(c("edge-edge1", "edge-edge2")),
    c("edge1", "edge2")
  )
  expect_equal(from_g6_edge_id("edge1"), "edge1")
  expect_equal(from_g6_edge_id(""), "")
  expect_equal(from_g6_edge_id("edge-"), "")
  expect_equal(from_g6_edge_id("edge-edge-test"), "edge-test")
})

test_that("to_g6_combo_id works correctly", {
  expect_equal(to_g6_combo_id("combo1"), "combo-combo1")
  expect_equal(
    to_g6_combo_id(c("combo1", "combo2")),
    c("combo-combo1", "combo-combo2")
  )
  expect_equal(to_g6_combo_id(character(0)), character(0))
  expect_equal(to_g6_combo_id(NULL), NULL)
  expect_equal(to_g6_combo_id(""), "combo-")
  expect_equal(to_g6_combo_id(789), "combo-789")
})

test_that("from_g6_combo_id works correctly", {
  expect_equal(from_g6_combo_id("combo-combo1"), "combo1")

  expect_equal(
    from_g6_combo_id(c("combo-combo1", "combo-combo2")),
    c("combo1", "combo2")
  )

  expect_equal(from_g6_combo_id("combo1"), "combo1")
  expect_equal(from_g6_combo_id(""), "")
  expect_equal(from_g6_combo_id("combo-"), "")
  expect_equal(from_g6_combo_id("combo-combo-test"), "combo-test")
})

test_that("to_g6_port_id works correctly", {
  expect_equal(to_g6_port_id("in1", "node1"), "node1-in1")
  expect_equal(
    to_g6_port_id(c("in1", "out1"), "node1"),
    c("node1-in1", "node1-out1")
  )
  expect_equal(
    to_g6_port_id("in1", c("node1", "node2")),
    c("node1-in1", "node2-in1")
  )
  expect_equal(to_g6_port_id(character(0), "node1"), character(0))
  expect_equal(to_g6_port_id(NULL, "node1"), NULL)
  expect_equal(to_g6_port_id("in1", ""), "-in1")
})

test_that("from_g6_port_id works correctly", {
  expect_equal(from_g6_port_id("node1-in1", "node1"), "in1")
  expect_equal(
    from_g6_port_id(c("node1-in1", "node1-out1"), "node1"),
    c("in1", "out1")
  )
  expect_equal(from_g6_port_id("node2-in2", "node2"), "in2")
  expect_equal(from_g6_port_id("in1", "node1"), "in1")
  expect_equal(from_g6_port_id("", "node1"), "")
  expect_equal(from_g6_port_id("node1-", "node1"), "")
})

test_that("roundtrip conversions work correctly", {
  original_nodes <- c("node1", "node2", "node3")
  expect_equal(from_g6_node_id(to_g6_node_id(original_nodes)), original_nodes)

  original_edges <- c("edge1", "edge2", "edge3")
  expect_equal(from_g6_edge_id(to_g6_edge_id(original_edges)), original_edges)

  original_combos <- c("combo1", "combo2", "combo3")
  expect_equal(
    from_g6_combo_id(to_g6_combo_id(original_combos)),
    original_combos
  )
})

test_that("g6_from_graph works", {
  g <- new_graph(
    nodes = list(
      list(id = 1, style = list(labelText = "Node 1")),
      list(id = 2, style = list(labelText = "Node 2"))
    ),
    edges = list(
      list(
        source = 1,
        target = 2,
        style = list(
          labelText = "Edge from 1 to 2"
        )
      )
    )
  )
  res <- g6_from_graph(g)
  expect_s3_class(res, c("g6", "htmlwidget"))
})

test_that("is_variadic_block works", {
  expect_false(is_variadic_block(new_dataset_block()))
  expect_true(is_variadic_block(new_rbind_block()))
})

test_that("create_block ports works", {
  blk <- new_scatter_block()
  ports <- create_block_ports(blk, "test")
  expect_s3_class(ports, "g6_ports")
  expect_length(ports, 2)
  expect_identical(ports[[1]]$type, "input")
  expect_identical(ports[[2]]$type, "output")
})

test_that("resolve_target_ports works", {
  blocks <- as_blocks(c(
    a = new_dataset_block(),
    b = new_head_block(),
    c = new_rbind_block()
  ))
  links <- as_links(
    c(
      new_link("a", "b", input = "data"),
      new_link("a", "c"),
      new_link("b", "c")
    )
  )
  res <- resolve_target_ports(links, blocks)
  expect_identical(res[[1]], "node-b-data")
  expect_identical(res[[2]], "node-c-in")
  expect_identical(res[[3]], "node-c-in")
})

library(testthat)

test_that("get_children_from_links works correctly", {
  # Create sample links
  links <- list(
    from = c("a", "a", "b"),
    to = c("b", "c", "d"),
    input = c("data", "data", "data")
  )
  class(links) <- c("links", "list")

  children <- get_children_from_links(links) # Check that node-a has two children
  expect_true("node-a" %in% names(children))
  expect_length(children[["node-a"]], 2)
  expect_true("node-b" %in% children[["node-a"]])
  expect_true("node-c" %in% children[["node-a"]])

  # Check that node-b has one child
  expect_true("node-b" %in% names(children))
  expect_length(children[["node-b"]], 1)
  expect_equal(children[["node-b"]], list("node-d"))

  # Check that node-c and node-d have no children
  expect_false("node-c" %in% names(children))
  expect_false("node-d" %in% names(children))
})

test_that("get_children_from_links handles empty links", {
  children <- get_children_from_links(list())
  expect_length(children, 0)
})

test_that("get_children_from_links handles duplicate targets", {
  # A node pointing to same target multiple times should only list it once
  links <- list(
    from = c("a", "a", "a"),
    to = c("b", "b", "c"),
    input = c("data1", "data2", "data3")
  )
  class(links) <- c("links", "list")

  children <- get_children_from_links(links)
  expect_length(children[["node-a"]], 2)
  expect_true("node-b" %in% children[["node-a"]])
  expect_true("node-c" %in% children[["node-a"]])
})

test_that("g6_nodes_from_blocks includes children when provided", {
  # Create simple blocks
  blocks <- list(
    a = new_dataset_block(),
    b = new_head_block()
  )
  class(blocks) <- c("blocks", "list")

  stacks <- list()
  class(stacks) <- c("stacks", "list")

  # Create children mapping (as list like get_children_from_links returns)
  children <- list(
    "node-a" = list("node-b")
  )

  nodes <- g6_nodes_from_blocks(blocks, stacks, children)

  # Check that node-a has children
  expect_length(nodes, 2)
  expect_equal(nodes[[1]]$id, "node-a")
  expect_equal(nodes[[1]]$children, list("node-b"))

  # Check that node-b has no children
  expect_equal(nodes[[2]]$id, "node-b")
  expect_null(nodes[[2]]$children)
})

test_that("g6_data_from_board computes and includes children", {
  # Create a simple board with blocks and links
  blocks <- as_blocks(c(
    a = new_dataset_block(),
    b = new_head_block(),
    c = new_head_block()
  ))

  links <- as_links(c(
    new_link("a", "b", input = "data"),
    new_link("b", "c", input = "data")
  ))

  stacks <- list()
  class(stacks) <- c("stacks", "list")

  board <- structure(
    list(
      blocks = blocks,
      links = links,
      stacks = stacks
    ),
    class = c("board", "list")
  )

  graph <- g6_data_from_board(board)

  # Check that nodes have children
  expect_length(graph$nodes, 3)

  # node-a should have node-b as child
  node_a <- graph$nodes[[1]]
  expect_equal(node_a$id, "node-a")
  expect_equal(node_a$children, list("node-b"))

  # node-b should have node-c as child
  node_b <- graph$nodes[[2]]
  expect_equal(node_b$id, "node-b")
  expect_equal(node_b$children, list("node-c"))

  # node-c should have no children
  node_c <- graph$nodes[[3]]
  expect_equal(node_c$id, "node-c")
  expect_null(node_c$children)
})

#' Toolbar item functions
#'
#' Functions for creating and working with toolbar
#' items for the DAG interface.
#'
#' @param id Unique identifier for the toolbar item.
#' @param icon Name of an icon to show in the toolbar.
#' @param js JavaScript code to execute when the entry is selected.
#' @param tooltip Optional tooltip text for the entry.
#' @param action Action to perform when the entry is selected.
#' @param x Object to test or extract toolbar items from.
#'
#' @details
#' \describe{
#'   \item{`new_toolbar_item()`}{Creates a new toolbar
#' item with the specified id, icon, JavaScript code, action function,
#' and tooltip text.}
#'   \item{`is_toolbar_item()`}{
#' Tests whether an object is a valid toolbar item.}
#'   \item{`toolbar_items()`}{Generic function to
#' extract toolbar items from various
#' objects like dock extensions, boards, or lists.}
#' }
#'
#' The `toolbar_items.dag_extension()` method provides
#' the following actions:
#' \itemize{
#'   \item Zoom in - Increases the graph zoom level.
#'   \item Zoom out - Decreases the graph zoom level.
#'   \item Auto fit - Automatically fits the entire graph within the viewport.
#'   \item Layout - Reapplies the graph layout algorithm to reorganize nodes.
#'   \item Add block - Opens interface to add a new block to the workflow.
#'   \item Add stack - Creates a new workflow stack.
#'   \item Remove selected - Removes currently selected elements from the graph.
#' }
#'
#' @rdname tool
#' @export
#' @return
#' \describe{
#'   \item{`new_toolbar_item()`}{A toolbar
#' item object of class "toolbar_item" containing
#' action and js functions, with id, icon, and tooltip attributes.}
#'   \item{`is_toolbar_item()`}{`TRUE` if `x` is
#' a toolbar item, `FALSE` otherwise.}
#'   \item{`toolbar_items()`}{A list of toolbar
#' items for the given object.}
#' }
new_toolbar_item <- function(id, icon, js, action = NULL, tooltip = NULL) {
  if (is_string(js)) {
    js_string <- js
    js <- function(...) js_string
  }

  stopifnot(
    is_string(id),
    is_string(icon),
    is.null(action) || is.function(action),
    is.function(js),
    is.null(tooltip) || is_string(tooltip)
  )

  if (is.null(tooltip)) {
    tooltip <- blockr.core::id_to_sentence_case(id)
  }

  structure(
    list(action = action, js = js),
    id = id,
    icon = icon,
    tooltip = tooltip,
    class = "toolbar_item"
  )
}

#' @param x Object
#' @rdname tool
#' @export
is_toolbar_item <- function(x) {
  inherits(x, "toolbar_item")
}

validate_toolbar_items <- function(x) {
  stopifnot(
    is.list(x),
    all(lgl_ply(x, is_toolbar_item)),
    anyDuplicated(chr_ply(x, toolbar_item_id)) == 0L
  )

  invisible(x)
}

toolbar_item_id <- function(x) attr(x, "id")

toolbar_item_icon <- function(x) attr(x, "icon")

toolbar_item_tooltip <- function(x) attr(x, "tooltip")

toolbar_item_action <- function(x, actions, session = get_session()) {

  if (!is_toolbar_item(x)) {

    res <- lapply(
      validate_toolbar_items(x),
      toolbar_item_action,
      actions,
      session
    )

    return(invisible(res))
  }

  stopifnot(is_toolbar_item(x))

  fun <- x[["action"]]

  if (is.null(fun)) {
    return(invisible(NULL))
  }

  res <- fun(actions, session)

  if (!inherits(res, "Observer")) {
    blockr_abort(
      "Expecting toolbar item server {toolbar_item_id(x)} to return an ",
      "observer.",
      class = "toolbar_item_return_invalid"
    )
  }

  invisible(res)
}

toolbar_item_js <- function(x, ns = NULL) {
  if (!is_toolbar_item(x)) {
    validate_toolbar_items(x)

    res <- paste(
      chr_ply(x, toolbar_item_js, ns = ns),
      collapse = " else "
    )

    return(
      paste0("(value, target, current) => {\n", res, "\n}")
    )
  }

  if (is.null(ns)) {
    ns <- NS(NULL)
  }

  paste0(
    "if (value === '",
    toolbar_item_id(x),
    "') {\n(",
    x[["js"]](ns),
    ")(value, target, current)\n}"
  )
}

build_toolbar <- function(x, ...) {
  if (!is_toolbar_item(x)) {
    validate_toolbar_items(x)

    res <- paste(
      chr_ply(x, build_toolbar, ...),
      collapse = ",\n"
    )

    return(
      paste0("( ) => [\n", res, "\n]")
    )
  }

  sprintf(
    "{ id : '%s' , value : '%s', title: '%s' }",
    toolbar_item_icon(x),
    toolbar_item_id(x),
    toolbar_item_tooltip(x)
  )
}

#' @rdname tool
#' @export
toolbar_items <- function(x) {
  UseMethod("toolbar_items")
}

#' @export
toolbar_items.dock_extension <- function(x) {
  list()
}

#' @export
toolbar_items.list <- function(x) {
  res <- lapply(x, toolbar_items)

  for (x in res) {
    validate_toolbar_items(x)
  }

  coal(unlst(res), list())
}

#' @export
toolbar_items.dock_extensions <- function(x) {
  res <- toolbar_items(as.list(x))
  validate_toolbar_items(res)
  res
}

#' @export
toolbar_items.dock_board <- function(x) {
  toolbar_items(blockr.dock::dock_extensions(x))
}

test_that("toolbar default", {
  # Extension
  ext <- new_dag_extension()
  tool <- toolbar_items(ext)

  expect_named(tool, NULL)
  expect_type(tool, "list")

  tst <- build_toolbar(tool)
  expect_type(tst, "character")

  item <- tool[[1]]
  expect_type(item, "list")
})

#' DAG extension
#'
#' Visualizes the DAG (directed acyclic graph) underlying a board and provides
#' UI elements to manipulate the board.
#'
#' @param graph A `graph` object (or `NULL`).
#' @param ... Forwarded to [blockr.dock::new_dock_extension()].
#'
#' @return A `dag_extension` object that extends the dock extension system
#' for visualizing and manipulating DAG workflows.
#' @rdname dag
#' @export
new_dag_extension <- function(graph = NULL, ...) {
  blockr.dock::new_dock_extension(
    dag_ext_srv(graph),
    dag_ext_ui,
    name = "Workflow",
    class = "dag_extension",
    ...
  )
}

#' @export
context_menu_items.dag_extension <- function(x) {
  list(
    new_context_menu_entry(
      name = "Create link",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            if (current.id === undefined) return;
            Shiny.setInputValue('%s', current.id.replace(/^node-/, ''), {priority: 'event'});
          }",
          ns("ctx_add_link")
        )
      },
      action = update_action_trigger(
        action_name = "add_link_action",
        input_name = "ctx_add_link"
      ),
      condition = function(board, target) target$type == "node",
      id = "create_link"
    ),
    new_context_menu_entry(
      name = "Remove block",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            if (current.id === undefined) return;
            Shiny.setInputValue('%s', current.id.replace(/^node-/, ''));
          }",
          ns("ctx_remove_block")
        )
      },
      action = update_action_trigger(
        action_name = "remove_block_action",
        input_name = "ctx_remove_block"
      ),
      condition = function(board, target) target$type == "node",
      id = "remove_block"
    ),
    new_context_menu_entry(
      name = "Remove link",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            if (current.id === undefined) return;
            Shiny.setInputValue('%s', current.id.replace(/^edge-/, ''));
          }",
          ns("ctx_remove_link")
        )
      },
      action = update_action_trigger(
        action_name = "remove_link_action",
        input_name = "ctx_remove_link"
      ),
      condition = function(board, target) target$type == "edge",
      id = "remove_link"
    ),
    new_context_menu_entry(
      name = "Append block",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            Shiny.setInputValue(
              '%s',
              current.id.replace(/^node-/, ''),
              {priority: 'event'}
            );
          }",
          ns("ctx_append_block")
        )
      },
      action = update_action_trigger(
        action_name = "append_block_action",
        input_name = "ctx_append_block"
      ),
      condition = function(board, target) target$type == "node",
      id = "append_block"
    ),
    new_context_menu_entry(
      name = "Add block",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            Shiny.setInputValue('%s', true, {priority: 'event'});
          }",
          ns("ctx_add_block")
        )
      },
      action = update_action_trigger(
        action_name = "add_block_action",
        input_name = "ctx_add_block"
      ),
      condition = function(board, target) target$type == "canvas",
      id = "add_block"
    ),
    new_context_menu_entry(
      name = "Create stack",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            Shiny.setInputValue('%s', true, {priority: 'event'});
          }",
          ns("ctx_create_stack")
        )
      },
      action = update_action_trigger(
        action_name = "add_stack_action",
        input_name = "ctx_create_stack"
      ),
      condition = function(board, target) target$type == "canvas",
      id = "create_stack"
    ),
    new_context_menu_entry(
      name = "Remove stack",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            if (current.id === undefined) return;
            Shiny.setInputValue('%s', current.id.replace(/^combo-/, ''));
          }",
          ns("ctx_remove_stack")
        )
      },
      action = update_action_trigger(
        action_name = "remove_stack_action",
        input_name = "ctx_remove_stack"
      ),
      condition = function(board, target) target$type == "combo",
      id = "remove_stack"
    ),
    new_context_menu_entry(
      name = "Edit stack",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            if (current.id === undefined) return;
            Shiny.setInputValue(
              '%s',
              current.id.replace(/^combo-/, ''),
              {priority: 'event'}
            );
          }",
          ns("ctx_edit_stack")
        )
      },
      action = update_action_trigger(
        action_name = "edit_stack_action",
        input_name = "ctx_edit_stack"
      ),
      condition = function(board, target) target$type == "combo",
      id = "edit_stack"
    ),
    new_context_menu_entry(
      name = "Copy",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            Shiny.setInputValue('%s', true, {priority: 'event'});
          }",
          ns("ctx_copy")
        )
      },
      action = update_action_trigger(
        action_name = "copy_selected_action",
        input_name = "ctx_copy"
      ),
      condition = function(board, target) {
        target$type %in% c("node", "combo")
      },
      id = "copy"
    ),
    new_context_menu_entry(
      name = "Cut",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            Shiny.setInputValue('%s', true, {priority: 'event'});
          }",
          ns("ctx_cut")
        )
      },
      action = update_action_trigger(
        action_name = "cut_selected_action",
        input_name = "ctx_cut"
      ),
      condition = function(board, target) {
        target$type %in% c("node", "combo")
      },
      id = "cut"
    ),
    new_context_menu_entry(
      name = "Paste",
      js = function(ns) {
        sprintf(
          "async (value, target, current) => {
            try {
              const text = await navigator.clipboard.readText();
              const data = JSON.parse(text);
              if (data && data.object === 'subboard') {
                Shiny.setInputValue('%s', text, {priority: 'event'});
              }
            } catch (err) {}
          }",
          ns("ctx_paste")
        )
      },
      action = update_action_trigger(
        action_name = "paste_action",
        input_name = "ctx_paste"
      ),
      condition = function(board, target) target$type == "canvas",
      id = "paste"
    )
  )
}

#' @export
toolbar_items.dag_extension <- function(x) {
  list(
    new_toolbar_item(
      id = "zoom_in",
      icon = "zoom-in",
      js = "(value, target, current) => {
        const graph = HTMLWidgets.find(
          `#${target.closest('.g6').id}`
        ).getWidget();
        graph.zoomTo(graph.getZoom() + 0.1);
      }"
    ),
    new_toolbar_item(
      id = "zoom_out",
      icon = "zoom-out",
      js = "(value, target, current) => {
        const graph = HTMLWidgets.find(
          `#${target.closest('.g6').id}`
        ).getWidget();
        graph.zoomTo (graph.getZoom() - 0.1);
      }"
    ),
    new_toolbar_item(
      id = "auto_fit",
      icon = "auto-fit",
      js = "(value, target, current) => {
        const graph = HTMLWidgets.find(
          `#${target.closest('.g6').id}`
        ).getWidget();
        graph.fitView();
      }"
    ),
    new_toolbar_item(
      id = "layout",
      icon = "reset",
      js = "(value, target, current) => {
        const graph = HTMLWidgets.find(
          `#${target.closest('.g6').id}`
        ).getWidget();
        graph.layout();
      }"
    ),
    new_toolbar_item(
      id = "add_block",
      icon = "icon-roundaddfill",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            Shiny.setInputValue('%s', true, {priority: 'event'});
          }",
          ns("tool_add_block")
        )
      },
      action = update_action_trigger(
        action_name = "add_block_action",
        input_name = "tool_add_block"
      )
    ),
    new_toolbar_item(
      id = "add_stack",
      icon = "icon-cascades",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            Shiny.setInputValue('%s', true, {priority: 'event'});
          }",
          ns("tool_add_stack")
        )
      },
      action = update_action_trigger(
        action_name = "add_stack_action",
        input_name = "tool_add_stack"
      )
    ),
    new_toolbar_item(
      id = "remove_selected",
      icon = "icon-delete",
      js = function(ns) {
        sprintf(
          "(value, target, current) => {
            Shiny.setInputValue('%s', true, {priority: 'event'});
          }",
          ns("tool_rm_selected")
        )
      },
      action = update_action_trigger(
        action_name = "remove_selected_action",
        input_name = "tool_rm_selected"
      )
    )
  )
}

#' @importFrom blockr.dock extension_block_callback
#' @export
extension_block_callback.dag_extension <- function(x, ...) {
  function(
    id,
    board,
    update,
    conditions,
    dag_extension,
    ...,
    session = get_session()
  ) {
    n_cnd <- reactive(
      sum(lengths(conditions()$error)),
      label = "n_cnd"
    )

    badge_count <- reactiveVal(0L)

    graph_ready <- reactive(
      isTRUE(dag_extension$proxy$session$input[[paste0(graph_id(), "-initialized")]]),
      label = "graph_ready"
    )

    observeEvent(
      req(graph_ready(), n_cnd() > 0L, n_cnd() != badge_count()),
      {
        n <- n_cnd()

        badge <- list(
          text = "",
          placement = "right-bottom",
          backgroundFill = "#dc2626",
          backgroundWidth = 10,
          backgroundHeight = 10
        )

        node_config <- list(
          list(
            id = to_g6_node_id(id),
            style = list(
              badges = list(badge)
            )
          )
        )

        g6_update_nodes(dag_extension$proxy, node_config)
        badge_count(n)
      },
      label = "show_error_badge"
    )

    observeEvent(
      req(graph_ready(), n_cnd() == 0L, badge_count() > 0L),
      {
        node_config <- list(
          list(
            id = to_g6_node_id(id),
            style = list(
              badges = list()
            )
          )
        )

        g6_update_nodes(dag_extension$proxy, node_config)
        badge_count(0L)
      },
      label = "clear_error_badge"
    )

    NULL
  }
}

library(shinytest2)

create_new_block <- function(app, name, id) {
  app$wait_for_idle()
  do.call(
    app$set_inputs,
    setNames(list(name), "board-add_block_action-add_block_selection")
  )
  app$wait_for_idle()
  app$click(selector = "#board-add_block_action-block-advanced-toggle")
  app$wait_for_idle()

  do.call(
    app$set_inputs,
    setNames(list(id), "board-add_block_action-add_block_id")
  )
  app$wait_for_idle()
  app$click("board-add_block_action-add_block_confirm")
  app$wait_for_idle()
}

create_new_link <- function(app, target, id) {
  app$wait_for_idle()
  do.call(
    app$set_inputs,
    setNames(list(target), "board-add_link_action-create_link")
  )

  app$click(selector = "#board-add_link_action-link-advanced-toggle")

  do.call(
    app$set_inputs,
    setNames(list(id), "board-add_link_action-add_link_id")
  )

  app$click("board-add_link_action-add_link_confirm")
  app$wait_for_idle()
}

right_click <- function(target, app, init = FALSE) {
  # Idea: we target element, get its bounding box, calculate center coordinates
  # and click there with right mouse button. We play with
  # Chromote API ...

  # Do not initialise variables more than once, Chromote
  # does not like it.
  x_result <- if (init) {
    app$get_chromote_session()$Runtime$evaluate(
      expression = sprintf(
        "let target = document.querySelector('%s');
        let rect = target.getBoundingClientRect();
        rect.left + rect.width / 2;",
        target
      )
    )
  } else {
    app$get_chromote_session()$Runtime$evaluate(
      expression = sprintf(
        "target = document.querySelector('%s');
        rect = target.getBoundingClientRect();
        rect.left + rect.width / 2;",
        target
      )
    )
  }

  # Get y coordinate
  y_result <- app$get_chromote_session()$Runtime$evaluate(
    expression = sprintf(
      "rect.top + rect.height / 2;"
    )
  )

  x_coord <- x_result$result$value
  y_coord <- y_result$result$value

  app$get_chromote_session()$Input$dispatchMouseEvent(
    type = "mousePressed",
    x = x_coord,
    y = y_coord,
    button = "right",
    clickCount = 1
  )

  app$get_chromote_session()$Input$dispatchMouseEvent(
    type = "mouseReleased",
    x = x_coord,
    y = y_coord,
    button = "right",
    clickCount = 1
  )

  app$wait_for_idle()
}

expect_values <- function(app, ...) {
  vals <- app$get_values()
  filter_names <- function(x) {
    grep("board-dag_extension", names(x), value = TRUE)
  }
  app$expect_values(
    ...,
    export = vals$export,
    input = filter_names(vals$input),
    output = filter_names(vals$output),
    transform = function(lines) {
      gsub("nonce=[0-9a-f]+", "nonce=<scrubbed>", lines, perl = TRUE)
    }
  )
}

test_that("sample_app works", {
  appdir <- system.file(package = "blockr.dag", "examples/empty")

  # when shinytest2 0.5.0 lands ...
  #local_app_support(appdir)

  app <- AppDriver$new(
    appdir,
    name = "empty-app",
    seed = 4323
  )
  expect_values(app)

  # Add a new block: with custom id
  app$wait_for_idle()
  app$click(selector = ".g6-toolbar-item[value=\"add_block\"")
  create_new_block(app, "dataset_block", "super_data_block")
  expect_values(app)

  # Append: show canvas context menu
  # Right-click on canvas to show context menu
  right_click("#board-dag_extension-graph", app, init = TRUE)
  app$click(selector = ".g6-contextmenu-li[value=\"add_block\"]")
  create_new_block(app, "head_block", "super_head_block")
  expect_values(app)

  # Right click on dataset block + add link with head block
  right_click("g#node-super_data_block", app)
  app$click(selector = ".g6-contextmenu-li[value=\"create_link\"]")
  create_new_link(app, "super_head_block", "super_link")
  expect_values(app)

  # Select new block and remove
  app$run_js(
    "HTMLWidgets
      .find('#board-dag_extension-graph')
      .getWidget()
      .setElementState('node-super_data_block', 'selected', false);"
  )
  app$set_inputs(
    `board-dag_extension-graph-selected_node` = "node-super_data_block",
    allow_no_input_binding_ = TRUE
  )
  app$run_js("$(\".g6-contextmenu\").show()")

  # Remove element
  app$click(selector = ".g6-toolbar-item[value=\"remove_selected\"")
  expect_values(app)

  app$stop()
})

# In reference to: https://github.com/BristolMyersSquibb/blockr.dag/issues/90
test_that("variadic app works", {
  appdir <- system.file(package = "blockr.dag", "examples/variadic")

  # when shinytest2 0.5.0 lands ...
  #local_app_support(appdir)

  app <- AppDriver$new(
    appdir,
    name = "variadic-app",
    seed = 4323
  )

  expect_values(app)
  app$wait_for_idle()
  app$stop()
})

---
title: "Context Menu Items"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Context Menu Items}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

```{r setup}
library(blockr.dag)
```

The __context menu__ provides __actions__ based on what element is __right-clicked__ in the graph.

## Available Actions

| Target Element | Available Actions |
|----------------|------------------|
| Canvas | • Add block<br>• Create stack |
| Node (block) | • Create link<br>• Remove block<br>• Append block |
| Edge (link) | • Remove link |
| Combo (stack) | • Remove stack<br>• Edit stack: add/remove blocks and change color |

## Creating Context Menu Entries

Context menu items are created with `new_context_menu_entry()`:

```{r}
# Context menu entry
entry <- new_context_menu_entry(
  name = "Custom Action",
  js = function(ns) "alert('Custom action')",
  action = function(board, update, ...) {
    # Server-side logic
  },
  condition = function(board, target) {
    target$type == "node"
  }
)
```

`js` can be a string representing a JavaScript function or an R function that takes the namespace `ns` as input and returns a string:

```r
function(ns) {
  sprintf(
    "(value, target, current) => {
      Shiny.setInputValue('%s', true, {priority: 'event'});
    }",
    ns("add_block")
  )
}
```

`condition` is an either a boolean or an R function that takes the `board` and `target` (the clicked element) as input and returns `TRUE` or `FALSE` to determine if the entry should be shown.

Any new entry has to go inside `context_menu_items.dag_extension` for registration:

```{r}
# Add custom context menu items to an extension
context_menu_items.dag_extension <- function(x) {
  list(
    new_context_menu_entry(
      name = "My Action",
      js = function(ns) {
        sprintf("Shiny.setInputValue('%s', true)", ns("my_action"))
      },
      action = my_custom_action("my_action"),
      condition = function(board, target) TRUE
    )
  )
}
```

---
title: "Toolbar Items"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Toolbar Items}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

```{r setup}
library(blockr.dag)
```

The __toolbar__ provides quick access to common graph operations.

## Available Actions

The toolbar includes several default operations:

- **Zoom in/out** - Navigate large workflows.
- **Auto fit** - Fit entire graph to view.
- **Layout** - Reset graph layout.
- **Add block** - Trigger create new blocks.
- **Add stack** - Trigger create stack.
- **Remove selected** - Delete selected elements (can be stacks, blocks and links)

## Creating Toolbar Items

Toolbar items are declared via `new_toolbar_item()`:

```{r}
# Toolbar item
item <- new_toolbar_item(
  id = "custom_tool",
  icon = "icon-custom",
  js = "console.log('Custom tool clicked')",
  action = function(board, update, ...) {
    # Server-side logic
  }
)
```

`js` can be a string representing a JavaScript function or an R function that takes the namespace `ns` as input and returns a string:

```r
function(ns) {
  sprintf(
    "(value, target, current) => {
      Shiny.setInputValue('%s', true, {priority: 'event'});
    }",
    ns("add_block")
  )
}
```

Any new item has to go inside `toolbar_items.dag_extension` for registration:

```r
toolbar_items.dag_extension <- function(x) {
  list(
    item
  )
```

---
title: "blockr.dag"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{blockr.dag}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

```{r setup}
library(blockr.dag)
library(blockr.core)
library(blockr.dock)
```

## Basic Usage

### Starting a DAG Board

We create an empty board from [blockr.dock](https://github.com/BristolMyersSquibb/blockr.dock) and serve it with the `new_dag_extension()` extension:

```{r}
# Create an empty board.
serve(new_dock_board(extensions = new_dag_extension()))
```

### Start from a graph object

You can also start with a custom graph structure:

```{r}
graph <- new_graph(
  nodes = list(
    list(id = 1),
    list(id = 2)
  ),
  edges = list(
    list(
      source = 1,
      target = 2,
      style = list(labelText = "1-2")
    )
  )
)

serve(
  new_dock_board(
    extensions = new_dag_extension(graph)
  )
)
```

## User Interface Elements

### Keyboard Shortcuts

- **Shift + Drag**: Create connections between blocks.
- **Ctrl + Backspace**: Remove selected elements.
- **Alt/Option(Mac) + Drag**: Multi-select elements.

## Developer Guide

### Where does blockr.dag fit in?

In the blockr [ecosystem](https://bristolmyerssquibb.github.io/blockr/), blockr.dag is a high level extension that usually comes at the end of the stack. It depends on blockr.core for core functionalities and blockr.dock for the user interface. It also relies on external packages like g6R for rendering and interaction, as well as extra block packages like blockr.dplyr and blockr.ggplot for additional block types:

```{mermaid}
graph TD
    subgraph Core Layer
        blockr.core[blockr.core]
    end

    subgraph Extension Blocks
        blockr.dplyr[blockr.dplyr]
        blockr.ggplot[blockr.ggplot]
    end

    subgraph UI Layer
        blockr.dock[blockr.dock]
        blockr.dag[blockr.dag]
    end

    subgraph External Dependencies
        dockViewR[dockViewR]
        g6R[g6R]
    end

    blockr.core --> blockr.dock
    blockr.dock --> blockr.dag
    dockViewR --> blockr.dock
    g6R --> blockr.dag

    blockr.core --> blockr.dplyr
    blockr.core --> blockr.ggplot

    blockr.dplyr -.-> blockr.dock
    blockr.dplyr -.-> blockr.dag
    blockr.ggplot -.-> blockr.dock
    blockr.ggplot -.-> blockr.dag
```

### Maintaining the DAG state

Server-side observers handle state synchronization between the block.core __board__ and the g6R output. `update` and `board` are inherited from `blockr.core`:

`update` is composed of:

| Category | Operation | Description |
|----------|-----------|-------------|
| `blocks` | `add` | New blocks to add (blocks object) |
|          | `mod` | Modified blocks (blocks object) |
|          | `rm` | Deleted blocks (block IDS, character) |
| `stacks` | `add` | New stacks to add (stacks object) |
|          | `mod` | Modified stacks (stacks object) |
|          | `rm` | Deleted stacks (stacks IDS, character) |
| `links`  | `add` | New links to add (links object) |
|          | `mod` | Modified links (links object) |
|          | `rm` | Deleted links (links IDS, character()) |

We listen to `update()` and apply the necessary changes to the `g6R` widget via its __proxy__ functions:

```{r}
# Update observer handles board changes
update_observer <- function(update, board, proxy) {
  observeEvent(update(), {
    upd <- update()

    if (length(upd$blocks$add)) {
      add_nodes(upd$blocks$add, board$board, proxy)
    }

    if (length(upd$stacks$mod)) {
      update_combos(upd$stacks$mod, board$board, proxy)
    }

    # Handle other update types...
  })
}
```

## Extending the Package

### Custom context menu and toolbar

You can learn more in the following vignettes:

- [Context menu](https://bristolmyerssquibb.github.io/blockr.dag/articles/context-menu.html).
- [Toolbar](https://bristolmyerssquibb.github.io/blockr.dag/articles/toolbar.html).

### Extension block callback

blockr.dock exposes a callback system to extend block functionality. You can learn more in the following [vignette](https://bristolmyerssquibb.github.io/blockr.dag/articles/block-callback.html).

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g6r.R
\name{g6_edges_from_links}
\alias{g6_edges_from_links}
\alias{g6_nodes_from_blocks}
\alias{g6_combos_data_from_stacks}
\alias{g6_data_from_board}
\title{Create network data from board}
\usage{
g6_edges_from_links(links, blocks)

g6_nodes_from_blocks(blocks, stacks, children = NULL)

g6_combos_data_from_stacks(stacks)

g6_data_from_board(board)
}
\arguments{
\item{links}{Board links.}

\item{blocks}{Board blocks.}

\item{stacks}{Board stacks.}

\item{children}{Named list of children node IDs (optional).}

\item{board}{Board object.}
}
\description{
Create network data from board
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g6r.R
\name{resolve_target_ports}
\alias{resolve_target_ports}
\title{Resolve target port IDs for links}
\usage{
resolve_target_ports(links, blocks)
}
\arguments{
\item{links}{Board links}

\item{blocks}{Board blocks}
}
\value{
Character vector of target port IDs
}
\description{
Resolve target port IDs for links
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph.R
\name{new_graph}
\alias{new_graph}
\alias{is_graph}
\alias{graph_nodes}
\alias{graph_edges}
\alias{graph_combos}
\alias{as_graph}
\alias{as_graph.graph}
\alias{as_graph.list}
\title{Graph object}
\usage{
new_graph(nodes = list(), edges = list(), combos = list())

is_graph(x)

graph_nodes(x)

graph_edges(x)

graph_combos(x)

as_graph(x, ...)

\method{as_graph}{graph}(x, ...)

\method{as_graph}{list}(x, ...)
}
\arguments{
\item{nodes}{Graph nodes (i.e. board blocks).}

\item{edges}{Graph edges (i.e. board links).}

\item{combos}{Node groups (i.e. board stacks).}

\item{x}{Object to test or convert.}

\item{...}{Generic consistency.}
}
\value{
\describe{
\item{\code{new_graph()}}{A graph object
of class "graph" containing nodes, edges, and combos lists.}
\item{\code{is_graph()}}{\code{TRUE} if \code{x} is a
graph object, \code{FALSE} otherwise.}
\item{\code{graph_nodes()}}{A list of graph nodes.}
\item{\code{graph_edges()}}{A list of graph edges.}
\item{\code{graph_combos()}}{A list of graph
combos (node groups).}
\item{\code{as_graph()}}{A graph object converted
from the input.}
}
}
\description{
R list-based representation of a g6 graph object.
}
\details{
\describe{
\item{\code{new_graph()}}{Creates a new graph
object with the specified nodes, edges, and combos.}
\item{\code{is_graph()}}{Tests whether an object
is a valid graph object.}
\item{\code{graph_nodes()}}{Extracts the nodes
component from a graph object.}
\item{\code{graph_edges()}}{Extracts the edges
component from a graph object.}
\item{\code{graph_combos()}}{Extracts the combos
component from a graph object.}
\item{\code{as_graph()}}{Generic function to
convert objects to graph format.}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tool.R
\name{new_toolbar_item}
\alias{new_toolbar_item}
\alias{is_toolbar_item}
\alias{toolbar_items}
\title{Toolbar item functions}
\usage{
new_toolbar_item(id, icon, js, action = NULL, tooltip = NULL)

is_toolbar_item(x)

toolbar_items(x)
}
\arguments{
\item{id}{Unique identifier for the toolbar item.}

\item{icon}{Name of an icon to show in the toolbar.}

\item{js}{JavaScript code to execute when the entry is selected.}

\item{action}{Action to perform when the entry is selected.}

\item{tooltip}{Optional tooltip text for the entry.}

\item{x}{Object}
}
\value{
\describe{
\item{\code{new_toolbar_item()}}{A toolbar
item object of class "toolbar_item" containing
action and js functions, with id, icon, and tooltip attributes.}
\item{\code{is_toolbar_item()}}{\code{TRUE} if \code{x} is
a toolbar item, \code{FALSE} otherwise.}
\item{\code{toolbar_items()}}{A list of toolbar
items for the given object.}
}
}
\description{
Functions for creating and working with toolbar
items for the DAG interface.
}
\details{
\describe{
\item{\code{new_toolbar_item()}}{Creates a new toolbar
item with the specified id, icon, JavaScript code, action function,
and tooltip text.}
\item{\code{is_toolbar_item()}}{
Tests whether an object is a valid toolbar item.}
\item{\code{toolbar_items()}}{Generic function to
extract toolbar items from various
objects like dock extensions, boards, or lists.}
}

The \code{toolbar_items.dag_extension()} method provides
the following actions:
\itemize{
\item Zoom in - Increases the graph zoom level.
\item Zoom out - Decreases the graph zoom level.
\item Auto fit - Automatically fits the entire graph within the viewport.
\item Layout - Reapplies the graph layout algorithm to reorganize nodes.
\item Add block - Opens interface to add a new block to the workflow.
\item Add stack - Creates a new workflow stack.
\item Remove selected - Removes currently selected elements from the graph.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g6r.R
\name{is_variadic_block}
\alias{is_variadic_block}
\title{Check if a block is variadic}
\usage{
is_variadic_block(block)
}
\arguments{
\item{block}{A block object}
}
\value{
Logical indicating if the block is variadic
}
\description{
A block is variadic if it has NA arity and no named inputs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ext.R
\name{new_dag_extension}
\alias{new_dag_extension}
\title{DAG extension}
\usage{
new_dag_extension(graph = NULL, ...)
}
\arguments{
\item{graph}{A \code{graph} object (or \code{NULL}).}

\item{...}{Forwarded to \code{\link[blockr.dock:extension]{blockr.dock::new_dock_extension()}}.}
}
\value{
A \code{dag_extension} object that extends the dock extension system
for visualizing and manipulating DAG workflows.
}
\description{
Visualizes the DAG (directed acyclic graph) underlying a board and provides
UI elements to manipulate the board.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g6r.R
\name{create_block_ports}
\alias{create_block_ports}
\title{Create block ports for g6 node}
\usage{
create_block_ports(block, id)
}
\arguments{
\item{block}{Block object.}

\item{id}{Block ID.}
}
\description{
Create block ports for g6 node
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g6r.R
\name{get_children_from_links}
\alias{get_children_from_links}
\title{Get children relationships from links}
\usage{
get_children_from_links(links)
}
\arguments{
\item{links}{Board links}
}
\value{
Named list where each element is a list of child node IDs (for JSON array conversion)
}
\description{
Get children relationships from links
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ctx.R
\name{new_context_menu_entry}
\alias{new_context_menu_entry}
\alias{is_context_menu_entry}
\alias{context_menu_items}
\title{Context menu functions}
\usage{
new_context_menu_entry(
  name,
  js,
  action = NULL,
  condition = TRUE,
  id = tolower(gsub(" +", "_", name))
)

is_context_menu_entry(x)

context_menu_items(x)
}
\arguments{
\item{name}{Name of the context menu entry.}

\item{js}{JavaScript code to execute when the entry is selected.}

\item{action}{Action to perform when the entry is selected.}

\item{condition}{Condition to determine if the entry should be shown.}

\item{id}{Unique identifier for the context menu entry.
Inferred from \code{name} if not provided}

\item{x}{Object}
}
\value{
\describe{
\item{\code{new_context_menu_entry()}}{A context menu
entry object of class "context_menu_entry" containing
condition, action, and js functions, with name and id attributes.}
\item{\code{is_context_menu_entry()}}{\code{TRUE} if \code{x} is
a context menu entry, \code{FALSE} otherwise.}
\item{\code{context_menu_items()}}{A list of context
menu items for the given object.}
}
}
\description{
Functions for creating and working with context
menu entries.
}
\details{
\describe{
\item{\code{new_context_menu_entry()}}{Creates a new context menu
entry with the specified name, JavaScript code, action function,
and display condition.}
\item{\code{is_context_menu_entry()}}{
Tests whether an object is a valid context menu entry.}
\item{\code{context_menu_items()}}{Generic function to
extract context menu items from various
objects like dock extensions, boards, or lists.}
}

The \code{context_menu_items.dag_extension()} method
provides the following actions:
\itemize{
\item Create link - Creates connections between workflow nodes.
\item Remove block - Removes individual blocks from the workflow.
\item Remove link - Removes connections between workflow nodes.
\item Append block - Adds a new block after the selected node.
\item Create stack - Creates a new workflow stack.
\item Remove stack - Removes an entire workflow stack.
\item Edit stack - Opens stack editing interface.
\item Add block - Adds a new block to the canvas.
}
}
